package org.alltimeflashdreamer.dimdom;

import java.io.*;
import java.net.URL;
import java.util.*;
import org.enhydra.xml.io.*;
import org.enhydra.xml.xmlc.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;
import org.xml.sax.*;

/**
 * Title: AbstractXHTMLDocument<br>
 * Description: This class wraps the necessary functionality around the "logic"
 * that is implemented in the DimDom engine, reference implementation. <br>
 * Each subclass of this class must perform extended, not referencial logic
 * operations using the specified SiteData instance of this class.<br>
 * The ideal usage of a subclass would be to instantiate the fully filled
 * SiteData object (e.g. as a subclass) and pass it to the DimDom, e.g. from
 * within a servlet: <font size="-1"><pre>
 * //fill every information into the SiteData within a subclass' constructor
 * MySiteData mySiteData = new MySiteData();
 * //call your own subclass' print method (this class knows which file it shall use)
 * response.getWriter().write(DimDom.print(mySiteData));
 * </pre></font>
 *
 * @author    Copyright (c) 2001/02 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

public abstract class AbstractXHTMLDocument implements XHTMLDocument
{
    /**
     * pointer to AbstractXHTMLSiteData object which is used to process the
     * dynamic data within the reference implementation
     */
    protected AbstractXHTMLSiteData siteData;
    /** the list of template nodes having an "id" attribute */
    protected Hashtable hIDNodeList;
    /**
     * the list of nodes and their repeated child elements (micro-template
     * structure nodes)
     */
    protected Hashtable hMicroTemplateNodes = new Hashtable();
    /**
     * the list of already processed id nodes, to find out whether it's a
     * repeated command on the same node or not, stored together with the number
     * of repetitions; this number is altered on each repetition of the node
     */
    protected Hashtable hProcessedIDs = new Hashtable();

    /** the org.w3c.dom.Document instance of the template DOM tree */
    private Document originalDoc;
    /** the list of table rows that have an id tag (must be handled separately) */
    private Hashtable hRepeatedTableRows = new Hashtable();
    /**
     * time stamp of the template file's save date; if the source is a file,
     * this value is > 0 after the parsing, else it remains 0.
     */
    private long lLastModified = 0;


//----------------- XHTMLDocument interface methods -------------------------//
    /**
     * Forces the XHTMLDocument instance to print out its dynamically adapted
     * content as a fully qualified XHTML String.<br>
     * This is done after the cloning of the internal Document pointer and the
     * calling of the processDynamicData method, which updates the DOM.
     *
     * @return                     the full XHTML string
     * @exception IOException      thrown by the parser (input source)
     * @exception DimDomException  thrown by the processor, if something is/went
     *      wrong
     */
    public String print()
        throws IOException, DimDomException
    {
        //deep-clone document node and fill in dynamic data
        Document clonedDoc = ( Document )originalDoc.cloneNode( true );
        processDynamicData( clonedDoc );

        //print DOMFormatters output into a StringWriter object
        StringWriter stringWriter = new StringWriter();
        DOMFormatter domFormatter = new DOMFormatter();
        OutputOptions options = domFormatter.getDefaultOutputOptions( clonedDoc );
        Formatter formatter = domFormatter.getFormatter( clonedDoc, options, false );
        formatter.write( clonedDoc, stringWriter );

        //clean-up: clear list of processed id's
        hProcessedIDs.clear();

        //return the fully qualified and modified XHTML string
        return stringWriter.toString();
    }

    /**
     * This method is (typically) called by the servlet instantiating a specific
     * subclass of this class that needs to be filled with dynamic data.<br>
     * It is defined within a separate interface to hide the functionality of
     * implementing classes to the servlet programmer, because the updating
     * process is uni-directional (thread safety!).
     *
     * @param siteData             an instance of a class implementing
     *      org.alltimeflashdreamer.dimdom.XHTMLSiteData interface
     * @exception SAXException     thrown by the parser
     * @exception IOException      thrown by the parser (input source)
     * @exception DimDomException  thrown by the processor, if something is/went
     *      wrong
     */
    public void setSiteData( XHTMLSiteData siteData )
        throws SAXException, IOException, DimDomException
    {
        this.siteData = ( AbstractXHTMLSiteData )siteData;

        //the parse methods prepare fast-find Hashtable filled with the "ID nodes" too!
        if ( this.siteData.isFileSource() )
            this.parseFile( siteData.getTemplateFileName() );
        else if ( this.siteData.isStringSource() )
            this.parseString( siteData.getString() );
        else
            this.parseInputStream( siteData.getURL().openStream() );
    }

    /**
     * This method can be called be extending classes to directly access an
     * Element specified by the id attribute value within the template. The
     * implementation here searches the internal list of id nodes for the one
     * with the given id. If no such node is found, it's assumed that this is a
     * common attribute setting, and a node having the same <i>name</i> is
     * searched and returned, if found, else null is returned.
     *
     * @param hIDNodeList          where the found nodes are stored in
     * @param sElementIDValue      the id value of the node (or the name, in
     *      case of the setting of a common attribute)
     * @param clonedDoc            the Document object to search for a common
     *      node template, in case
     * @return                     the Element from the list or a common node or
     *      null
     * @exception DimDomException  thrown if the node to be processed is not a
     *      Element node
     */
    protected Element getElement( Hashtable hIDNodeList,
            String sElementIDValue, Document clonedDoc )
        throws DimDomException
    {
        Element returnValue = ( Element )hIDNodeList.get( sElementIDValue.toLowerCase() );
        try
        {
            //this element may be null, if not found (see "else")
            if ( returnValue == null )
                returnValue = ( Element )clonedDoc.getElementsByTagName( sElementIDValue ).item( 0 );
        } catch ( ClassCastException cce )
        {
            throw new DimDomException( "Node to process was not an org.w3c.dom.Element node! " +
                    "Please make sure you don't try to identify not-Element nodes as " +
                    "DimDom nodes in the template.", cce );
        } catch ( Exception e )
        {
            //ignore a missing common template node (can come to here only in that case)
        }

        return returnValue;
    }

    /**
     * Processes one node of the template having an id attribute value, along
     * the specifications of DimDom. The executed tasks are:
     * <ul>
     *   <li> The visibility of the node is checked; if invisible, nothing else
     *   is to be done (node may be replaced by something)</li>
     *   <li> The embedding in a link is done, if necessary</lI>
     *   <li> The list of defined attributes is processed (common and individual
     *   attributes)</li>
     *   <li> The list of defined values is processed (single value,
     *   one-dimensional or two-dimensional list of values, with selected
     *   entry/ies or without)</li>
     * </ul>
     *
     *
     * @param dimdomNode          the object containing the dynamic information;
     *      must not be null!
     * @param element             the element to which the dynamic information
     *      shall be applied
     * @param bIsRepeatedCommand  tells the engine about whether this node has
     *      already been handled within the "getNextRepeatedElement
     *      intelligence" or not; if not, the method has to take care on its own
     *      about what to do with repeated data <i>structures</i> (not nodes)
     * @throws DimDomException    if some error occurs while adding the
     *      information to the node
     */
    protected void processDimDomNode( DimDomNodeImpl dimdomNode,
            Element element, boolean bIsRepeatedCommand )
        throws DimDomException
    {
        //element isn't allowed to be null!
        if ( element == null )
            throw new DimDomException( "The element with id \"" + dimdomNode.getNodeIdentifier() +
                    " was null!", new NullPointerException() );

        //process invisibility, if necessary, and return when done
        if ( !dimdomNode.isVisible() )
        {
            String sReplacement = dimdomNode.getReplacingText();
            Node replacingNode;
            if ( sReplacement.equals( "" ) )
                replacingNode = createEmptyTextNode( element.getOwnerDocument() );
            else
                replacingNode = element.getOwnerDocument().createTextNode( sReplacement );
            hide( element, replacingNode );
            return;
        }

        //process link embedding, if necessary
        if ( dimdomNode.hasEmbeddingLink() )
            wrapWithLink( element, dimdomNode.getEmbeddingLink() );

        //process attributes: common, simple and child attributes are recognized
        Element childElement = null;
        while ( dimdomNode.hasMoreAttributeObjects() )
        {
            DimDomNodeImpl.AttributeObject nextAttributeObject =
                    ( DimDomNodeImpl.AttributeObject )dimdomNode.getNextAttributeObject();
            if ( nextAttributeObject.isCommon() )
                setCommonAttribute( element, nextAttributeObject.getAttr() );
            else if ( nextAttributeObject.isChildAttribute() )
            {
                if ( childElement == null )
                    childElement = createAndAddNextDimDomChild( element, bIsRepeatedCommand );
                setAttribute( childElement, nextAttributeObject.getAttr() );
            }
            else
                setAttribute( element, nextAttributeObject.getAttr() );
        }

        //process values: single value -> just set the text; value list -> setContent method,
        //as long as there are more values, if it's a list node, else collect the values and
        //call the setTexts method afterwards; value field: same as value list, but a whole
        //row may be selected
        Vector vNotListValues = null;
        boolean bIsDimDomListNode = isDimDomListNode( element );
        while ( dimdomNode.hasMoreValueObjects() )
        {
            DimDomNodeImpl.ValueObject nextValueObject =
                    ( DimDomNodeImpl.ValueObject )dimdomNode.getNextValueObject();
            if ( nextValueObject.isSingleValue() )
            {
                //single value

                if ( bIsDimDomListNode )
                    setContent( dimdomNode.getNodeIdentifier(),
                            createAndAddNextDimDomChild( element, bIsRepeatedCommand ),
                            nextValueObject.getNextValue(),
                            nextValueObject.hasSelection(),
                            bIsRepeatedCommand );
                else
                    setText( element, nextValueObject.getNextValue(), nextValueObject.hasSelection() );
            }
            else if ( nextValueObject.isValueList() )
            {
                //value list

                int iCtr = 0;
                vNotListValues = ( ( bIsDimDomListNode ) ? null : new Vector() );
                while ( nextValueObject.hasMoreValues() )
                {
                    if ( bIsDimDomListNode )
                        setContent( dimdomNode.getNodeIdentifier(),
                                createAndAddNextDimDomChild( element, bIsRepeatedCommand ),
                                nextValueObject.getNextValue(),
                                ( nextValueObject.getSelectedRow() == iCtr ),
                                bIsRepeatedCommand );
                    else
                        vNotListValues.add( nextValueObject.getNextValue() );
                    iCtr++;
                }
                if ( !bIsDimDomListNode )
                    setTexts( element, vNotListValues, nextValueObject.getSelectedRow() );
            }
            else
            {
                //value field
                int iCtr = 0;
                int iNEntriesPerRow = 0;
                vNotListValues = ( ( bIsDimDomListNode ) ? null : new Vector() );
                while ( nextValueObject.hasMoreValues() )
                {
                    String[] sNextValueRow = nextValueObject.getNextValueRow();
                    if ( bIsDimDomListNode )
                        setContent( dimdomNode.getNodeIdentifier(),
                                createAndAddNextDimDomChild( element, bIsRepeatedCommand ),
                                sNextValueRow,
                                ( nextValueObject.getSelectedRow() == iCtr ),
                                nextValueObject.getSelectedColumn(),
                                bIsRepeatedCommand );
                    else
                    {
                        if ( sNextValueRow != null )
                        {
                            iNEntriesPerRow = sNextValueRow.length;
                            for ( int i = 0; i < sNextValueRow.length; i++ )
                                vNotListValues.add( nextValueObject.getNextValue() );
                        }
                    }
                    iCtr++;
                }
                if ( !bIsDimDomListNode )
                    setTexts( element, vNotListValues, iNEntriesPerRow,
                            nextValueObject.getSelectedRow(), nextValueObject.getSelectedColumn() );
            }
        }
    }

    /**
     * This class doesn't implement the logic that is necessary to find out
     * which node is the next to be processed in micro-template structures.
     * Therefore, this method is to be implemented by extending classes. They
     * will need the getter (and perhaps the setter) method of the repeated
     * elements to get the list of RepeatedElement instances, the wrapper
     * classes for these "special" nodes.
     *
     * @param sNodeIdentifier      the id attribute value of the node to be
     *      processed
     * @param element              the current node to be processed, for
     *      deciding how to behave
     * @return                     the next repeated element, i.e. a clone that
     *      is already appended/inserted correctly
     * @exception DimDomException  wrapper for unexpected errors
     */
    protected abstract Element getNextRepeatedElement( String sNodeIdentifier, Element element )
        throws DimDomException;

    /**
     * Sets the given string content to the given node by setting the content to
     * the first found text node. If no such text node exists, a new one is
     * created, filled and added to element. If sText contains newline
     * characters ("\n"), they are replaced with a "br" node.
     *
     * @param element              the element to set the content to
     * @param sText                the text to set to element; if null or empty,
     *      the node's text is set to ""
     * @exception DimDomException  wrapped DOMException, in case
     */
    protected void setText( Element element, String sText )
        throws DimDomException
    {
        if ( sText == null )
            sText = "";
        try
        {
            //find first Text; if missing, create a new one
            Text firstText = getFirstText( element );

            //try updating the Text node; if this fails, remove the text node
            //and add a freshly created one, without
            try
            {
                StringTokenizer st = new StringTokenizer( sText, "\n" );
                if ( st.countTokens() == 0 )
                    firstText.setData( "" );
                else
                {
                    firstText.setData( st.nextToken() );
                    while ( st.hasMoreElements() )
                    {
                        Element br = firstText.getOwnerDocument().createElement( "br" );
                        firstText.getParentNode().appendChild( br );
                        Text nextLine = firstText.getOwnerDocument().createTextNode( st.nextToken() );
                        firstText.getParentNode().appendChild( nextLine );
                    }
                }
            } catch ( Exception e )
            {
                try
                {
                    firstText.getParentNode().removeChild( firstText );
                } catch ( Exception e2 )
                {
                }
                element.appendChild( element.getOwnerDocument().createTextNode( sText ) );
            }
        } catch ( DOMException de )
        {
            throw new DimDomException( "Couldn't set content \"" + sText +
                    "\" to the element! Make sure that the text is XHTML compliant.", de );
        }
    }

    /**
     * Searches the first text node done the tree, starting with element. If no
     * such node is found, a newly created, empty Text is returned, which is
     * appended to element.
     *
     * @param element  the root node to search
     * @return         a previously existing child Text or a new Text
     */
    protected Text getFirstText( Element element )
    {
        Text firstText = findFirstText( element );
        if ( firstText == null )
        {
            firstText = element.getOwnerDocument().createTextNode( "" );
            element.appendChild( firstText );
        }
        return firstText;
    }

    /**
     * Sets a list of string values to a node that is not a list node (see
     * isDimDomListNode method). If this situation occurs, the values are added
     * as a comma-separated list, and if the given index is a defined one in the
     * given vector, it's embedded in a "b" tag (HTML bold tag).
     *
     * @param element              node to add the list to
     * @param vNotListValues       list of values (only String instances)
     * @param iSelectedIndex       the selected index; if out of range (e.g.
     *      -1), no value is marked as bold
     * @exception DimDomException  in case something went wrong, this is an own
     *      or a wrapped DOMException
     */
    protected void setTexts( Element element, Vector vNotListValues, int iSelectedIndex )
        throws DimDomException
    {
        //get first text
        Text currentText = getFirstText( element );

        //append each value comma-separatedly; if it's the selected one,
        //embed it in a "b" tag, append the embedded tag and create a
        //new text node to be filled with the next value(s)
        for ( int i = 0; i < vNotListValues.size(); i++ )
        {
            String sNextValue = ( String )vNotListValues.get( i );
            if ( i > 0 )
                currentText.setData( ", " );
            if ( i == iSelectedIndex )
            {
                element.appendChild( createSelectedElement( element.getOwnerDocument(), sNextValue ) );
                currentText = element.getOwnerDocument().createTextNode( "" );
                element.appendChild( currentText );
            }
            else
                currentText.setData( currentText.getData() + sNextValue );
        }
    }

    /**
     * Sets a 2D field of string values to a node that is not a list node (see
     * isDimDomListNode method). If this situation occurs, the values are added
     * as a comma-separated list, and if the given row/column index is a
     * selected one, the related value is embedded in a "b" tag (HTML bold tag).
     * In case the selected row index is set, but the selected column index is
     * out of range, each entry of the selected row is marked as selected.
     *
     * @param element              node to add the list to
     * @param vNotListValues       list of values (only String instances)
     * @param iNColumnsPerRow      how many columns are there per row?
     * @param iSelectedRow         the selected row index; if out of range (e.g.
     *      -1), no value is marked as bold
     * @param iSelectedColumn      the selected column index; if out of range
     *      (e.g. -1), no value is marked as bold
     * @exception DimDomException  in case something went wrong, this is an own
     *      or a wrapped DOMException
     */
    protected void setTexts( Element element, Vector vNotListValues, int iNColumnsPerRow,
            int iSelectedRow, int iSelectedColumn )
        throws DimDomException
    {
        //get first text
        Text currentText = getFirstText( element );

        int iNRows = vNotListValues.size() / iNColumnsPerRow;
        boolean bWholeRowIsSelected = false;
        if ( ( iSelectedRow >= 0 ) && ( iSelectedColumn < 0 ) )
            bWholeRowIsSelected = true;

        for ( int i = 0; i < iNRows; i++ )
        {
            for ( int j = 0; j < iNColumnsPerRow; j++ )
            {
                String sNextValue = ( String )vNotListValues.get( i * iNColumnsPerRow + j );
                if ( !( ( i == 0 ) && ( j == 0 ) ) )
                    currentText.setData( ", " );

                //select entry in case the current row index equals the selected row
                //index AND the selected column index is out or range OR the current
                //column index is equal to the selected column index
                if ( ( i == iSelectedRow ) && ( ( bWholeRowIsSelected ) || ( j == iSelectedColumn ) ) )
                {
                    element.appendChild( createSelectedElement( element.getOwnerDocument(), sNextValue ) );
                    currentText = element.getOwnerDocument().createTextNode( "" );
                    element.appendChild( currentText );
                }
                else
                    currentText.setData( currentText.getData() + sNextValue );
            }
        }
    }

    /**
     * Creates a new node of type "b" and appends a text node to it, containing
     * sText. This node is then returned.
     *
     * @param doc               the Document to create nodes with
     * @param sText             the text to show selected; if null, "" is added
     * @return                  the selected node
     * @throws DimDomException  a wrapped DOMException, in case
     */
    protected Element createSelectedElement( Document doc, String sText )
        throws DimDomException
    {
        Element b = null;
        try
        {
            b = doc.createElement( "b" );
            setText( b, ( sText != null ) ? sText : "" );
        } catch ( DOMException de )
        {
            throw new DimDomException( "Couldn't create selected node or add text to it!", de );
        }
        return b;
    }

    /**
     * Sets the content of element to sContent. There are two scenarios for this
     * method:
     * <ul>
     *   <li> "Simple tags": This is done by finding all text and entity nodes
     *   and replacing them with a newly created one, having sContent as text
     *   data
     *   <li> "Repeatable tags": ol/ul, select, tr -> individual appending of
     *   new child node(s), as needed
     * </ul>
     * This is done in case element is not null only.<br>
     * Notes:
     * <ul>
     *   <li> This is the method that is called for two-dimensional value fields
     *   only! For one-dimensional value fields, see the other setContent
     *   method. </li>
     *   <li> The behaviour on null or empty values is always as it would be a
     *   "normal" value, <i>except</i> for the select element: An option
     *   element's value attribute is removed and the disabled attribute is set
     *   (HTML 4.0+)</li>
     * </ul>
     *
     *
     * @param sNodeIdentifier      the name of the currently processed
     *      DimDomNode
     * @param element              the Element to set the content to
     * @param sContent             the content (if null, it's set to "")
     * @param bIsSelected          whether to select this element or not
     * @param bIsRepeatedCommand   tells the method whether this is a
     *      micro-template structure node or not
     * @exception DimDomException  if element is null, or a wrapped
     *      DOMException, in case
     */
    protected void setContent( String sNodeIdentifier, Element element,
            String sContent, boolean bIsSelected, boolean bIsRepeatedCommand )
        throws DimDomException
    {
        //check element and content
        if ( element == null )
            throw new DimDomException( "Element to set content \"" + sContent + "\" to was null!" );
        if ( sContent == null )
            sContent = "";

        //clone element, if necessary, and use appropriate setText/setTexts method to fill element
        if ( isHTMLListNode( element ) )
            setText( element, sContent, bIsSelected );
        else if ( element.getNodeName().equalsIgnoreCase( "option" ) )
        {
            element.setAttribute( "value", sContent );
            setText( element, sContent );
            setOptionTagSelected( element, bIsSelected );
        }
        else if ( isTableRowNode( element ) )
        {
            //get the "cursor position" (how many repetitions?), starting with 0
            int iTdThRepetitionCtr = getProcessedRepetitionsCount( sNodeIdentifier );

            //get number of td/th elements in the template table row; for the (very rare)
            //case that this value should be null (quite strange micro-template structure),
            //recalculate the value; if it's still null then, throw a DimDomException
            int iNTdTh = getStoredTableDataChildCount( sNodeIdentifier );
            if ( iNTdTh == 0 )
                iNTdTh = getTableDataChildCount( element );
            if ( iNTdTh == 0 )
                throw new DimDomException( "Currently processed child node of DimDomNode \"" +
                        sNodeIdentifier + "\" doesn't contain child elements of type td or th!" );

            //calculate the index (the cell number within the current row)
            int iIndex = ( iTdThRepetitionCtr + 1 ) % iNTdTh;

            //in case a new table row must be started (which is true if the index is 0 [again]
            //AND the given id doesn't represent a micro-template node), clone the element and append
            //the clone to the element's parent node and replace the element pointer
            if ( ( iIndex == 0 ) && ( !bIsRepeatedCommand ) )
            {
                Element clonedTableRow = ( Element )element.cloneNode( true );
                clonedTableRow.removeAttribute( "id" );
                element.getParentNode().appendChild( clonedTableRow );
                element = clonedTableRow;
            }

            //find current td/th subnode, then find the deepest element of this subnode
            //and set the content to this deepest element
            int iCtr = 0;
            for ( Node childOfTR = element.getFirstChild();
                    childOfTR != null; childOfTR = childOfTR.getNextSibling() )
            {
                if ( isTableDataNode( childOfTR ) )
                {
                    //comparison based on 0, not on 1!
                    if ( iCtr == iIndex )
                    {
                        Element currentThTd = findFirstLeafElement( childOfTR );
                        setText( currentThTd, sContent, bIsSelected );
                        break;
                    }
                    iCtr++;
                }
            }
        }
        else
            setText( element, sContent, bIsSelected );
    }

    /**
     * Sets the children's content of element using the value list parameter.
     * There are two scenarios for this method:
     * <ul>
     *   <li> "Simple tags": This is done by finding all text and entity nodes
     *   and replacing them with a newly created one, having a "non-list list"
     *   as text data (see setTexts method)
     *   <li> "Repeatable tags": ol/ul, select, tr -> individual appending of
     *   new child node(s): for ol/ul, each li tag receives a non-list list; for
     *   select, the first value is set to the value attribute of the option
     *   parameter, each further value is set to the option tags content, as a
     *   non-list list, and if the boolean parameter is true, the (option)
     *   element is selected anyway; for tr each td/th child receives a value,
     *   if the amount of values is less or equal to the amount of td/th
     *   children, else the last td/th child receives a non-list list
     * </ul>
     * This is done only in case both element and sValues aren't null AND
     * sValues.length is at least 1!<br>
     * Notes:
     * <ul>
     *   <li> This is the method that is called for two-dimensional value fields
     *   only! For one-dimensional value fields, see the other setContent
     *   method. </li>
     *   <li> The behaviour on null or empty values is always as it would be a
     *   "normal" value, <i>except</i> for the select element: An option
     *   element's value attribute is removed and the disabled attribute is set
     *   (HTML 4.0+)</li>
     * </ul>
     *
     *
     * @param sNodeIdentifier      the name of the currently processed
     *      DimDomNode
     * @param element              the Element to set the content to
     * @param sValues              the content (if null, it's set to "")
     * @param bIsSelectedRow       whether it's a selected row or not
     * @param iSelectedColumn      if it's selected row and the selected column
     *      is a valid index, only this column entry will be selected, else the
     *      whole row; if it's not the selected row, this parameter is ignored
     * @param bIsRepeatedCommand   tells the method whether this is a
     *      micro-template structure node or not
     * @exception DimDomException  if element or sValues are null, or a wrapped
     *      DOMException, in case
     */
    protected void setContent( String sNodeIdentifier, Element element, String[] sValues,
            boolean bIsSelectedRow, int iSelectedColumn, boolean bIsRepeatedCommand )
        throws DimDomException
    {
        //check element and content
        if ( element == null )
            throw new DimDomException( "Element to set content the value list to was null!" );
        if ( ( sValues == null ) || ( sValues.length == 0 ) )
            return;
        //throw new DimDomException( "Value list to set to the element was null or empty!" );

        //use appropriate setText/setTexts method to fill element (which is already a
        //newly created and appended clone!)
        Vector vValues = new Vector();
        for ( int i = 0; i < sValues.length; i++ )
            vValues.add( sValues[i] );

        if ( isHTMLListNode( element ) )
            //ol/ul?
            setTexts( element, vValues, iSelectedColumn );
        else if ( element.getNodeName().equalsIgnoreCase( "option" ) )
        {
            if ( ( sValues[0] != null ) && !( sValues[0].equals( "" ) ) )
                element.setAttribute( "value", sValues[0] );
            else
            {
                element.removeAttribute( "value" );
                element.setAttribute( "disabled", "on" );
            }
            if ( sValues.length == 1 )
                setText( element, sValues[0] );
            else
            {
                vValues.remove( 0 );
                //was used for the value attribute's value
                setTexts( element, vValues, iSelectedColumn );
            }
            setOptionTagSelected( element, bIsSelectedRow );
        }
        else if ( isTableRowNode( element ) )
        {
            //each time this method is called, a new table row must be started (if it's
            //necessary): clone the element, append the clone to the element's
            //parent node and replace the element pointer
            if ( !bIsRepeatedCommand )
            {
                Element clonedTableRow = ( Element )element.cloneNode( true );
                clonedTableRow.removeAttribute( "id" );
                element.getParentNode().appendChild( clonedTableRow );
                element = clonedTableRow;
            }

            //iterate through table data child nodes and fill them with the values
            int iCtr = 0;
            for ( Node child = element.getFirstChild(); child != null; child = child.getNextSibling() )
            {
                if ( isTableDataNode( child ) )
                {
                    boolean bIsSelectedElement = ( ( bIsSelectedRow ) &&
                            ( ( iSelectedColumn == iCtr ) || ( iSelectedColumn < 0 ) ) );

                    //several cases are possible :
                    //- has same amount of values as table data node children: no problem;
                    //- has less table data node children: set not-list list to last such child node
                    //- has more table data node children:
                    //    set those empty that can't receive a value anymore (see below, after the loop)
                    Element firstLeaf = findFirstLeafElement( child );
                    if ( !vValues.isEmpty() )
                        setText( firstLeaf, ( String )vValues.remove( 0 ), bIsSelectedElement );
                    else
                        firstLeaf.getParentNode().replaceChild(
                                createEmptyTextNode( element.getOwnerDocument() ), firstLeaf );
                    iCtr++;
                }
            }

            //in case there are more values than td/th child nodes, append a not-list list to the last one
            if ( !vValues.isEmpty() )
            {
                //find last td/th child and update its content (don't forget the selection!)
                for ( Node child = element.getLastChild(); child != null; child = child.getPreviousSibling() )
                {
                    if ( isTableDataNode( child ) )
                    {
                        Element tdth = findFirstLeafElement( child );
                        vValues.add( 0, findFirstText( tdth ).getData() );
                        iCtr--;
                        setTexts( tdth, vValues, ( iSelectedColumn - iCtr ) );
                        break;
                    }
                }
            }
        }
        else
            setTexts( element, vValues, sValues.length, ( ( bIsSelectedRow ) ? 0 : -1 ), iSelectedColumn );
    }

    /**
     * Sets or unsets the selected attribute of the option tag, depending on the
     * boolean parameter.
     *
     * @param option       the HTML option tag to update
     * @param bIsSelected  set selected?
     */
    protected final void setOptionTagSelected( Element option, boolean bIsSelected )
    {
        if ( bIsSelected )
            option.setAttribute( "selected", "on" );
        else
            option.removeAttribute( "selected" );
    }

    /**
     * Just calls setText(element, sText), if bIsSelected is false; else a
     * selected element is created and appended to element.
     *
     * @param element              the node to set the text to
     * @param sText                the text to set to the node
     * @param bIsSelected          whether to mark the text as selected or not
     * @exception DimDomException  in case something went wrong, this is an own
     *      or a wrapped DOMException
     */
    protected void setText( Element element, String sText, boolean bIsSelected )
        throws DimDomException
    {
        if ( !bIsSelected )
            setText( element, sText );
        else
        {
            Text previousText = findFirstText( element );
            if ( previousText != null )
                previousText.getParentNode().removeChild( previousText );
            element.appendChild( createSelectedElement( element.getOwnerDocument(), sText ) );
        }
    }

    /**
     * Sets the attribute and its value to <i>each</i> node found within the
     * document that has the same node name as element. Be careful using this
     * command, it is designed to e.g. replace each font face attribute within a
     * document.<br>
     * Because this method needs a template element to know the name of the
     * common attribute node(s), the parameter element must not be null, but an
     * existing template node for the setting of common attributes
     *
     * @param element              the element having the common node id (may
     *      not be null)
     * @param attribute            the Attr instance to add to these nodes
     * @exception DimDomException  wrapped DOMException, in case
     */
    protected void setCommonAttribute( Element element, Attr attribute )
        throws DimDomException
    {
        if ( element == null )
            throw new DimDomException( "The template element for the setting of common attributes must not be null!" );

        NodeList commonNodes = element.getOwnerDocument().getElementsByTagName( element.getNodeName() );
        for ( int i = 0; i < commonNodes.getLength(); i++ )
            setAttribute( ( Element )commonNodes.item( i ), attribute );
    }

    /**
     * Parses the InputStream specified in stream (parsing is done any time the
     * method is executed).
     *
     * @param stream               the InputStream to parse from
     * @exception SAXException     thrown by the parser
     * @exception IOException      thrown by the parser (input source)
     * @exception DimDomException  thrown by the processor, if something is/went
     *      wrong
     */
    protected void parseInputStream( InputStream stream )
        throws SAXException, IOException, DimDomException
    {
        lLastModified = 0;
        originalDoc = new XHTMLParser().parse( stream );
        hIDNodeList = buildListOfIDNodes( originalDoc );
    }

    /**
     * Parses the InputStream specified in stream (parsing is done any time the
     * method is executed).
     *
     * @param sInput               the String to parse
     * @exception SAXException     thrown by the parser
     * @exception IOException      thrown by the parser (input source)
     * @exception DimDomException  thrown by the processor, if something is/went
     *      wrong
     */
    protected void parseString( String sInput )
        throws SAXException, IOException, DimDomException
    {
        lLastModified = 0;
        originalDoc = new XHTMLParser().parse( sInput );
        hIDNodeList = buildListOfIDNodes( originalDoc );
    }

    /**
     * Parses the file specified in sFileName, but only during initialization or
     * if the specified file has a new last modification date.
     *
     * @param sFileName            the name of the file to be parsed
     * @exception SAXException     thrown by the parser
     * @exception IOException      thrown by the parser (input source)
     * @exception DimDomException  thrown by the processor, if something is/went
     *      wrong
     */
    protected void parseFile( String sFileName )
        throws SAXException, IOException, DimDomException
    {
        parseFile( new File( sFileName ) );
    }

    /**
     * Parses the parameter file, but only during initialization or if the
     * specified file has a new last modification date.
     *
     * @param file                 the file to parse
     * @exception SAXException     thrown by the parser
     * @exception IOException      thrown by the parser (input source)
     * @exception DimDomException  thrown by the processor, if something is/went
     *      wrong
     */
    protected void parseFile( File file )
        throws SAXException, IOException, DimDomException
    {
        //update only when check is passed
        if ( !file.exists() )
            throw new FileNotFoundException( "File doesn't exist!" );

        //for debugging: first load?
        boolean bIsFirstLoad = ( originalDoc == null );

        //read file only if originalDoc is null or file has a new last modification date (and update internal
        //last modification value and the id node list, in case)
        if ( ( bIsFirstLoad ) || ( lLastModified != file.lastModified() ) )
        {
            lLastModified = file.lastModified();
            originalDoc = new XHTMLParser().parse( file );
            hIDNodeList = buildListOfIDNodes( originalDoc );
        }
    }

    /**
     * Sets an attribute name and value to the parameter Element. <i>This method
     * </i> must be used by extending subclasses. If element or attribute is
     * null, an exception is thrown.
     *
     * @param element              the Element to set the attribute to
     * @param attribute            the Attr to add
     * @exception DimDomException  thrown if element or attribute is null
     */
    protected void setAttribute( Element element, Attr attribute )
        throws DimDomException
    {
        if ( attribute == null )
            throw new DimDomException( "Attribute to set was null!" );
        if ( element == null )
            throw new DimDomException( "Element to set attribute \"" + attribute.getNodeName() +
                    "=" + attribute.getNodeValue() + "\" was null!" );

        try
        {
            //DOM forbids to directly add the given Attr node, because it's created for
            //the original document node, so fake it; if it's the "checked" or the
			//"selected" attribute AND the value is null or "", then the attribute is removed
			boolean bRemoveAttribute = false;
			if ( isEmpty( attribute.getNodeValue() ) )
			{
				if ( ( attribute.getNodeName().equals( "selected" ) && element.getNodeName().equals( "option" ) ) ||
				     ( attribute.getNodeName().equals( "checked" ) && element.getNodeName().equals( "input" ) ) )
					bRemoveAttribute = true;
			}
			if ( bRemoveAttribute )
			    element.removeAttribute( attribute.getNodeName() );
			else
				element.setAttribute( attribute.getNodeName(), attribute.getNodeValue() );
        } catch ( DOMException de )
        {
            throw new DimDomException( "Attribute \"" + attribute.getNodeName() +
                    "=" + attribute.getNodeValue() + "\" couldn't be set!", de );
        }
    }

    /**
     * Wraps the parameter Element with a link specified by its target URL and
     * frame. <i>This method</i> must be used by extending subclasses.<br>
     * This is done only in case element is not null (if embeddingLink is null,
     * nothing is done).
     *
     * @param element              the Element to embed in a link
     * @param embeddingLink        the instance of DimDomNodeImpl describing the
     *      embedding "a" tag and its attributes
     * @exception DimDomException  if element is null or a wrapped DOMException,
     *      in case
     */
    protected void wrapWithLink( Element element, DimDomNodeImpl.EmbeddingLink embeddingLink )
        throws DimDomException
    {
        //checks
        if ( element == null )
            throw new DimDomException( "Element to wrap with link " +
                    embeddingLink.getHref().getNodeValue() + " having target " +
                    embeddingLink.getTarget().getNodeValue() + " was null!" );
        if ( embeddingLink == null )
            return;

        //create link node
        Element linkElement = element.getOwnerDocument().createElement( "a" );
        setAttribute( linkElement, embeddingLink.getHref() );
        if ( embeddingLink.getTarget() != null )
            setAttribute( linkElement, embeddingLink.getTarget() );

        //replacing must take place first, as long as element is still a child node of parentNode,
        //the appending of element to the new link node must happen afterwards
		if ( isTableDataNode( element ) || isTableRowNode( element ) )
		{
			Element firstLeaf = findFirstLeafElement( element );
			firstLeaf.getParentNode().replaceChild( linkElement, firstLeaf );
			linkElement.appendChild( firstLeaf );
		}
		else
		{
			element.getParentNode().replaceChild( linkElement, element );
			linkElement.appendChild( element );
		}
    }

    /**
     * Hides the Element parameter by replacing it with another node.<br>
     * If the element is a "tr" element, then the children of type "th" or "td"
     * are searched and their content is hided, and its height attribute is
     * removed. If element is a "td" or a "th" itself, then its children are
     * hided. Else the element itself is replaced with the empty text node
     * parameter.<br>
     * This is done only if both element and replacingNode aren't not null.
     *
     * @param element              element to hide or replace with a Text node
     * @param replacingNode        the node to replace element with
     * @exception DimDomException  if element or replacingNode is null or a
     *      wrapped DOMException, in case
     */
    protected void hide( Element element, Node replacingNode )
        throws DimDomException
    {
        //checks
        if ( element == null )
            throw new DimDomException( "Element to hide was null!" );
        if ( replacingNode == null )
            throw new DimDomException( "Replacing node for the element to hide was null!" );

        //in case element is a tr node, preserve the structure and hide the td and th
        //child nodes only
        if ( isTableRowNode( element ) )
        {
            element.removeAttribute( "height" );
            NodeList childNodes = element.getChildNodes();
            for ( int i = 0; i < childNodes.getLength(); i++ )
            {
                Node nextNode = childNodes.item( i );
                if ( ( nextNode instanceof Element ) && ( isTableDataNode( nextNode ) ) )
                {
                    Element nextChildElement = ( Element )nextNode;
                    hideChildElements( nextChildElement, replacingNode );
                }
            }
        }
        //in case element is a th or a td node, hide its child elements
        else if ( isTableDataNode( element ) )
            hideChildElements( element, replacingNode );
		//in case it's some other node that IS A CHILD of a table element,
		//remove it only, whithout replacing node
		else if ( element.getNodeName().equalsIgnoreCase( "table" ) )
		    element.getParentNode().removeChild( element );
        //else replace element with the replacing node
        else
            element.getParentNode().replaceChild( replacingNode, element );
    }

    /**
     * Builds a hashtable containing ID/XHTML element pairs to find the nodes
     * more easily when modifying them during the following update process with
     * the dynamic values.
     *
     * @param doc                  the org.w3c.dom.Document containing the DOM
     *      root node to be searched
     * @return                     where the found nodes are stored in after the
     *      search
     * @exception DimDomException  thrown if an id attribute value is found more
     *      than once in the template
     */
    protected Hashtable buildListOfIDNodes( Document doc )
        throws DimDomException
    {
        Hashtable hReturnValue = new Hashtable();
        findAndAddIDNodes( doc.getChildNodes(), hReturnValue );
        return hReturnValue;
    }

    /**
     * Creates an "empty text node", i.e. a text node having one child node (the
     * "&amp;nbsp;" entity) and no text content.
     *
     * @param doc  the owner document which shall created the node
     * @return     the "empty" node, which is actually a nbsp entity within a
     *      attributeless div node
     */
    protected Node createEmptyTextNode( Document doc )
    {
        Element returnValue = doc.createElement( "div" );
        EntityReference nbspEntity = doc.createEntityReference( "nbsp" );
        returnValue.appendChild( nbspEntity );

        return returnValue;
    }

    /**
     * Searches the tree down until he finds a leaf node of type Element. If
     * such a node is found, it's returned.
     *
     * @param node  the starting node for the search
     * @return      the (actually) deepest element
     */
    protected Element findFirstLeafElement( Node node )
    {
        Element returnValue = ( ( node instanceof Element ) ? ( Element )node : null );

        for ( Node child = node.getFirstChild(); child != null; child = child.getNextSibling() )
        {
            if ( child instanceof Element )
                returnValue = findFirstLeafElement( child );
        }

        return returnValue;
    }

    /**
     * Returns whether node is a DimDom list node or not, i.e. whether its name
     * is either "ol" or "ul" or "select" or "tr" - or none of them.
     *
     * @param node  the node to query
     * @return      whether it's a DimDom list node or not
     */
    protected boolean isDimDomListNode( Node node )
    {
        if ( isHTMLListNode( node ) ||
                node.getNodeName().equalsIgnoreCase( "select" ) ||
                isTableRowNode( node ) )
            return true;
        else
            return false;
    }

    /**
     * Returns whether node is a HTML list node or not, i.e. whether its name is
     * either "ol" or "ul" - or none of them.
     *
     * @param node  the node to query
     * @return      whether it's a HTML list node or not
     */
    protected boolean isHTMLListNode( Node node )
    {
        if ( node.getNodeName().equalsIgnoreCase( "ol" ) ||
                node.getNodeName().equalsIgnoreCase( "ul" ) )
            return true;
        else
            return false;
    }

    /**
     * Returns whether node is a table dent node or not, i.e. whether its name
     * is either "td" or "th" - or none of them.
     *
     * @param node  the node to be checked
     * @return      whether it's a table data node or not
     */
    protected boolean isTableDataNode( Node node )
    {
        if ( node.getNodeName().equalsIgnoreCase( "td" ) ||
                node.getNodeName().equalsIgnoreCase( "th" ) )
            return true;
        else
            return false;
    }

    /**
     * Returns whether node is a table row node or not, i.e. whether its name
     * is "tr" or not.
     *
     * @param node  the node to be checked
     * @return      whether it's a table row node or not
     */
    protected boolean isTableRowNode( Node node )
    {
        if ( node.getNodeName().equalsIgnoreCase( "tr" ) )
            return true;
        else
            return false;
    }

    /**
     * Returns whether the given element is a micro-template node or not. This
     * is always false if the given node is null, because this means that a
     * common attribute command is to be processed. It's false too in case the
     * internal list of micro-template nodes doesn't contain the given node,
     * else it's true. After this check, the internal list of micro-template
     * nodes is updated with the number of repetitions (starting with 1, to get
     * the correct value).
     *
     * @param sNodeIdentifier  the id attribute value of the node
     * @param element          the element to check
     * @return                 whether this node was processed earlier or not
     */
    protected boolean isRepeatedCommand( String sNodeIdentifier, Element element )
    {
        boolean bReturnValue = false;

        //common attribute nodes can never be repeated, i.e. part of a
        //micro-template structure within the template!
        if ( element == null )
            return bReturnValue;

        //check whether the given node is stored in
        //the list of micro-template nodes and adapt the counter list
        if ( hMicroTemplateNodes.get( sNodeIdentifier ) != null )
            bReturnValue = true;
        Integer numberOfRepetitions = ( Integer )hProcessedIDs.get( sNodeIdentifier );
        if ( numberOfRepetitions == null )
            hProcessedIDs.put( sNodeIdentifier, new Integer( 0 ) );
        else
            hProcessedIDs.put( sNodeIdentifier, new Integer( numberOfRepetitions.intValue() + 1 ) );

        return bReturnValue;
    }

    /**
     * Returns the total number of already processed repetitions of the node
     * having the given identifier. If sNodeIdentifier is null or "", 0 is
     * returned. 0 is returned too if the internal list of processed node
     * doesn't contain the given node identifier (yet), else the stored value is
     * returned.
     *
     * @param sNodeIdentifier  the value of the node's id attribute, i.e. the
     *      node identifier
     * @return                 0 or higher, the number of processed repetitions
     *      on the given node
     */
    protected int getProcessedRepetitionsCount( String sNodeIdentifier )
    {
        if ( ( sNodeIdentifier == null ) || ( sNodeIdentifier.equals( "" ) ) )
            return 0;

        Integer numberOfRepetitions = ( Integer )hProcessedIDs.get( sNodeIdentifier );
        return ( ( numberOfRepetitions == null ) ? 0 : numberOfRepetitions.intValue() );
    }

    /**
     * Searches the tree recursively, starting with node, for a node having the
     * given name. If such a node is found, it's returned, else null is returned
     *
     * @param node       root node to search
     * @param sNodeName  name to search for
     * @return           a node named sNodeName or null
     */
    protected Node findFirstNodeByName( Node node, String sNodeName )
    {
        if ( node.getNodeName().equals( sNodeName ) )
            return node;

        for ( Node child = node.getFirstChild(); child != null; child = child.getNextSibling() )
        {
            Node returnedNode = findFirstNodeByName( child, sNodeName );
            if ( returnedNode != null )
                return returnedNode;
        }

        return null;
    }

    /**
     * A new child element is created, appended and returned.<br>
     * Which node is chosen, is up to the specifications of the DimDom.
     * Supported node names are: form, frameset, select, ol/ul and tr.<br>
     * This method is used primarily to add child nodes in a "pseudo-repeated"
     * context, i.e. the programmer sent a bundle of values that must be filled
     * in by repeatedly calling this method and then processing the received
     * node. Another common usage is the adding of a hidden input to a form tag.
     * <br>
     * The second task of this method is to repeat some microtemplate structure,
     * e.g. a node that contains some subnodes needs to be repeated as a whole.
     * If so, the method clones the parameter element deeply and appends it to
     * the same parent node, at the <i>following</i> position. This behaviour
     * becomes active if the parameter element is not a supported node (as
     * listed above).
     *
     * @param element             element to which a new child shall be added
     * @param bIsRepeatedCommand  tells the method whether this is a
     *      micro-template structure node or not
     * @return                    the new child, as specified above
     * @throws DimDomException    a wrapped DOMException, in case
     */
    protected Element createAndAddNextDimDomChild(
            Element element, boolean bIsRepeatedCommand )
        throws DimDomException
    {
        if ( element == null )
            throw new DimDomException( "Element to add a new child to must not be null!" );

        Element nextListElement = null;

        //"select" creates a new option child tag
        if ( element.getNodeName().equalsIgnoreCase( "select" ) )
            nextListElement = element.getOwnerDocument().createElement( "option" );
        //"ol" or "ul" creates a new li child tag
        else if ( isHTMLListNode( element ) )
            nextListElement = element.getOwnerDocument().createElement( "li" );
        //"form" creates a new hidden input child tag
        else if ( element.getNodeName().equalsIgnoreCase( "form" ) )
        {
            nextListElement = element.getOwnerDocument().createElement( "input" );
            nextListElement.setAttribute( "type", "hidden" );
        }
        //"frameset" creates a new frame child tag
        else if ( element.getNodeName().equalsIgnoreCase( "frameset" ) )
            nextListElement = element.getOwnerDocument().createElement( "frame" );

        //first behaviour: append new child
        if ( nextListElement != null )
            element.appendChild( nextListElement );
        //second behaviour: insert clone after element (quite complicated...)
        else
        {
            //handle the following, very tricky case: if a micro-template node shall
            //be repeated AND this node is a tr node, then don't clone it, because
            //in this case, the tr node has been handled by the getNextRepeatedElement
            //intelligence, and this intelligence for sure did clone the node already
            if ( isTableRowNode( element ) && bIsRepeatedCommand )
                nextListElement = element;
            else
            {
                nextListElement = ( Element )element.cloneNode( true );
                element.getParentNode().appendChild( nextListElement );
            }
        }

        //return the next list element
        return nextListElement;
    }

    /**
     * This method recursively searches the given NodeList for each member of
     * the DOM tree having a defined "dimdom" attribute. When such a node is
     * found, a new instance of the internal RepeatedElement class is created
     * which stores a pointer to that node and knows how many times the node is
     * allowed to be repeated.<br>
     * Note: This method is visible for extending classes, because it is very
     * useful for the logic process, but it is declared final because it is
     * essential to work for <i>this</i> class as it is.
     *
     * @param nodes                the list of (child) nodes to search
     *      repeatedly
     * @param vRepeatedElements    the list to add the found repeated elements
     *      to
     * @exception DimDomException  thrown by the CommandParser, in case
     */
    protected final void findAndAddRepeatedElements( NodeList nodes, Vector vRepeatedElements )
        throws DimDomException
    {
        for ( int i = 0; i < nodes.getLength(); i++ )
        {
            //get current node and its attributes
            Node currentNode = nodes.item( i );
            NamedNodeMap attributes = currentNode.getAttributes();

            //if node's DIMDOM attribute is defined, add a new instance of RepeatedElement to the Vector
            if ( attributes != null )
            {
                //ignore nodes having a defined "id" attribute
                if ( attributes.getNamedItem( "id" ) != null )
                    continue;

                Node dimdomAttribute = attributes.getNamedItem( "dimdom" );
                if ( dimdomAttribute != null )
                    vRepeatedElements.add( new RepeatedElement( currentNode, dimdomAttribute.getNodeValue() ) );
            }

            //recursive call, but only, if no exception has occurred yet (no break until here)
            findAndAddRepeatedElements( currentNode.getChildNodes(), vRepeatedElements );
        }
    }

    /**
     * Returns the list of RepeatedElement instances that were initialized for
     * the given node or null, if no such nodes exist for the given node in the
     * template.
     *
     * @param sNodeIdentifier  the id attribute of the node to get the repeated
     *      elements for
     * @return                 the list of repeated elements or null
     */
    protected RepeatedElement[] getRepeatedElements( String sNodeIdentifier )
    {
        return ( RepeatedElement[] )hMicroTemplateNodes.get( sNodeIdentifier );
    }

    /**
     * Sets the list of RepeatedElement instances of the given node. If this
     * list is null, the element is removed from this list.<br>
     * Important: Calling this method changes the behaviour on these nodes
     * completely, because they are replaced with new ones!
     *
     * @param sNodeIdentifier   the id attribute of the node to get the repeated
     *      elements for
     * @param repeatedElements  the list of repeated elements or null
     */
    protected void setRepeatedElements(
            String sNodeIdentifier, RepeatedElement[] repeatedElements )
    {
        if ( repeatedElements == null )
            hMicroTemplateNodes.remove( sNodeIdentifier );
        else
            hMicroTemplateNodes.put( sNodeIdentifier, repeatedElements );
    }

    /**
     * Returns the <i>current</i> amount of td or th (see isTableDataNode
     * method) child elements for the given node identifier. This means that if
     * and only if the underlying node is a tr node (a table row to be repeated
     * as a whole), this method returns the amount of <i>first generation</i>
     * table data child elements of this tr node. In any other case, this method
     * returns 0.<br>
     * The sense of this method is to allow callers to inform themselves about
     * the situation within the underlying node's table row.<br>
     * Note: The "current" amount is a value that survives only a single
     * processing of a site. For each further dynamic filling of a site, the
     * value is recalculated.
     *
     * @param element  the Element for which the value shall be returned
     * @return         the number of "table data" (td/th) child nodes or 0
     */
    protected int getTableDataChildCount( Element element )
    {
        if ( !isTableRowNode( element ) )
            return 0;

        int iTdThCtr = 0;
        NodeList childNodes = element.getChildNodes();
        for ( int i = 0; i < childNodes.getLength(); i++ )
        {
            Node nextChildNode = childNodes.item( i );
            if ( isTableDataNode( nextChildNode ) )
            {
                //notice and...
                iTdThCtr++;

                //...remove child text nodes
                removeChildTextNodes( nextChildNode );
            }
        }
        return iTdThCtr;
    }

    /**
     * Returns the stored value for the amount of td/th child nodes, as
     * calculated int getTableDataChildCount.
     *
     * @param sNodeIdentifier  the id node for which the value shall be returned
     * @return                 the number of "table data" (td/th) child nodes or
     *      0
     */
    protected int getStoredTableDataChildCount( String sNodeIdentifier )
    {
        try
        {
            return ( ( Integer )hRepeatedTableRows.get( sNodeIdentifier ) ).intValue();
        } catch ( Exception e )
        {
            return 0;
        }
    }

//---------------------------- implementation specific methods ----------------------------//
    /**
     * This method is splits up the list of entries made by an addData method of
     * XHTMLSiteData. For each entry found, the method processEntry is called.
     * <br>
     * The parameter is a clone of the doc object to prevent thread safety and
     * to not modify the original object which represents a parsed file.
     * Implementing classes should be aware that the changed must be done on <i>
     * this</i> object, not on a newly instantiated one that is going to be
     * returned.
     *
     * @param clonedDoc            the org.w3c.dom.Document object of the
     *      template source
     * @exception DimDomException  if something happens during fill-in of the
     *      dynamic content
     */
    private void processDynamicData( Document clonedDoc )
        throws DimDomException
    {
        //checks against existency
        if ( ( this.siteData == null ) || ( this.siteData.getIdentifier() == null ) )
            return;

        //build a list of nodes from the CLONED document, not by cloning the
        //original id node list!
        Hashtable hClonedIDNodeList = buildListOfIDNodes( clonedDoc );

        //(re)build the list of micro-template nodes from the cloned id node list
        buildMicroTemplateNodeList( hClonedIDNodeList );

        //delete not needed subnodes of the template elements, e.g. text nodes
        clearTemplateElements( hClonedIDNodeList );

        while ( siteData.hasMoreDimDomNodeImpls() )
        {
            //prepare data from XHTMLSiteData
            DimDomNodeImpl nextDimDomNode = ( DimDomNodeImpl )this.siteData.getNextDimDomNode();

            //get original element and element to be processed (TBP)
            Element elem = getElement( hClonedIDNodeList, nextDimDomNode.getNodeIdentifier(), clonedDoc );

            //it's possible that a template node to set common attributes to is
            //null, because no such node exists in the template. if so, ignore the command
            if ( elem != null )
            {
                boolean bIsRepeatedCommand = isRepeatedCommand( nextDimDomNode.getNodeIdentifier(), elem );
                //if it's a repeated element, call the "logic" method first
                if ( bIsRepeatedCommand )
                    elem = getNextRepeatedElement( nextDimDomNode.getNodeIdentifier(), elem );

                //call processor on current element, if it's not repeated or the
                //repetition logic found a "next element"
                if ( elem != null )
                    processDimDomNode( nextDimDomNode, elem, bIsRepeatedCommand );
            }
        }

        //remove each node having a "dimdom=repeat(n)" attribute (because those
        //are the template nodes and have been cloned only)
        removeTemplateNodes( hClonedIDNodeList, clonedDoc );
    }

    /**
     * Hides each child element (!) of element, i.e. removes all of them and
     * adds the replacing node afterwards.
     *
     * @param element        the element to hide
     * @param replacingNode  the node to replace element with
     */
    protected void hideChildElements( Element element, Node replacingNode )
    {
        NodeList childNodes = element.getChildNodes();
        for ( int i = 0; i < childNodes.getLength(); i++ )
        {
            Node nextNode = childNodes.item( i );
            if ( nextNode instanceof Element )
            {
                Element nextChildElement = ( Element )childNodes.item( i );
                element.removeChild( nextChildElement );
            }
        }
        element.appendChild( replacingNode );
    }

    /**
     * Finds the first text down the tree, starting with node. The algorithm
     * recursively looks more than one level to search for text in font nodes,
     * etc.
     *
     * @param node  the starting node for the search
     * @return      the text node or null, if not found
     */
    protected final Text findFirstText( Node node )
    {
        if ( node instanceof Text )
            return ( Text )node;

        for ( Node child = node.getFirstChild(); child != null; child = child.getNextSibling() )
        {
            Text text = findFirstText( child );
            if ( text != null )
                return text;
        }

        return null;
    }

    /**
     * This method recursively searches the NodeList for each member of the DOM
     * tree for nodes having a defined ID attribute. When such a node is found,
     * it's added to the internal Hashtable, together with the ID value of the
     * node. Having this list, it's much easier to find a node again later.<br>
     * If a ID attribute is found having the same value like another (that was
     * already found and added to the list), a DimDomException informing the
     * user about the problem is stored and the cycle is broken up.<br>
     * Note: The id attribute values are stored in their lower case expression,
     * as specified for HTML, not XHTML!
     *
     * @param nodes                the next NodeList to search recursively
     * @param hIDNodeList          where the found nodes shall be stored in
     * @exception DimDomException  thrown if an id attribute value is found more
     *      than once in the template
     */
    private void findAndAddIDNodes( NodeList nodes, Hashtable hIDNodeList )
        throws DimDomException
    {
        for ( int i = 0; i < nodes.getLength(); i++ )
        {
            //get current node and its attributes
            Node currentNode = nodes.item( i );
            NamedNodeMap attributes = currentNode.getAttributes();

            //if node's ID attribute is defined, add it to the hashtable, together with the ID value
            if ( attributes != null )
            {
                Node idAttribute = attributes.getNamedItem( "id" );
                if ( ( idAttribute != null ) && ( idAttribute.getNodeName().equalsIgnoreCase( "id" ) ) )
                {
                    //is there already an entry for the current value? If yes, throw a DimDomException
                    if ( hIDNodeList.get( idAttribute.getNodeValue().toLowerCase() ) != null )
                        throw new DimDomException( "ID attribute value " + idAttribute.getNodeValue() +
                                " is used more than once in the template file!" );

                    hIDNodeList.put( idAttribute.getNodeValue().toLowerCase(), currentNode );
                }
            }

            //recursive call, but only, if no exception has occurred yet (no break until here)
            findAndAddIDNodes( currentNode.getChildNodes(), hIDNodeList );
        }
    }

    /**
     * This method is called after the very end of the processDynamicData method
     * and deletes each node of the (cloned and modified) DOM tree having a
     * "dimdom=repeat" attribute.<br>
     * After this, the previously stored table row (tr) nodes that were repeated
     * are deleted too.
     *
     * @param hIDNodeList          the list of id nodes to process
     * @param doc                  the Document which shall be cleaned up
     * @exception DimDomException  thrown by the command parser
     */
    private void removeTemplateNodes( Hashtable hIDNodeList, Document doc )
        throws DimDomException
    {
        //find and remove the repeated child elements of each id node
        Vector vElementsToDelete = new Vector();
        Enumeration eKeys = hIDNodeList.keys();
        while ( eKeys.hasMoreElements() )
        {
            Node currentNode = ( Node )hIDNodeList.get( eKeys.nextElement() );
            findAndAddRepeatedElements( currentNode.getChildNodes(), vElementsToDelete );
        }
        while ( !vElementsToDelete.isEmpty() )
        {
            Node currentNodeToDelete = ( ( RepeatedElement )vElementsToDelete.remove( 0 ) ).getRepeatedNode();
            currentNodeToDelete.getParentNode().removeChild( currentNodeToDelete );
        }

        //remove the repeated table row templates, which are the only tr nodes having an
        //id attribute, if necessary, by storing them in the list of nodes to be deleted
        if ( !hRepeatedTableRows.isEmpty() )
        {
            findAndAddTemplateTableRows( doc.getDocumentElement(), vElementsToDelete );
        }

		//remove those nodes that still have a repeated (dimdom) attribute, by storing
		//them in the list of nodes to be deleted
		findAndAddDimDomNodes( doc.getDocumentElement(), vElementsToDelete );

		//clear the list of elements marked to be deleted
        removeNodes( vElementsToDelete );
    }

    /**
     * Removes each node in the given list. If such a node doesn't have a parent
     * node, it's ignored.
     *
     * @param vNodesToDelete
     */
    protected final void removeNodes( Vector vNodesToDelete )
    {
        while ( !vNodesToDelete.isEmpty() )
        {
            Node currentNodeToDelete = ( Node )vNodesToDelete.remove( 0 );
            try
            {
                currentNodeToDelete.getParentNode().removeChild( currentNodeToDelete );
            } catch ( NullPointerException npe )
            {}
        }
    }

    /**
     * Searches the DOM tree for nodes named "tr" and having a defined "id"
     * attribute. If such a node is found, it's added to the given Vector. This
     * search is done <i>strictly case-insensitively</i> .
     *
     * @param node       the node to search
     * @param vElements  list o add found elements to
     */
    private void findAndAddTemplateTableRows( Node node, Vector vElements )
    {
        for ( Node child = node.getFirstChild(); child != null; child = child.getNextSibling() )
        {
            if ( isTableRowNode( child ) )
            {
                NamedNodeMap attrs = child.getAttributes();
                if ( attrs != null )
                {
                    for ( int i = 0; i < attrs.getLength(); i++ )
                    {
                        if ( attrs.item( i ).getNodeName().equalsIgnoreCase( "id" ) )
                        {
                            vElements.add( child );
                            break;
                        }
                    }
                }
            }

            //recursive call
            findAndAddTemplateTableRows( child, vElements );
        }
    }

    /**
     * Searches the DOM tree for nodes having a defined "dimdom"
     * attribute. If such a node is found, it's added to the given Vector. This
     * search is done <i>strictly case-sensitively</i> .
     *
     * @param node       the node to search
     * @param vElements  list to add found elements to
     */
	private void findAndAddDimDomNodes( Node node, Vector vElementsToDelete )
	{
		for ( Node child = node.getFirstChild(); child != null; child = child.getNextSibling() )
		{
			try {
				if ( child.getAttributes().getNamedItem( "dimdom" ) != null )
					vElementsToDelete.add( child );
			}
			catch (NullPointerException npe) {}

			findAndAddDimDomNodes( child, vElementsToDelete );
		}
	}

    /**
     * Checks each entry of the Hashtable: If there is a template element in a
     * "DimDom list" (i.e. a list that can have dynamically created subnodes),
     * it is deleted before the appending of the value-containing (new) child
     * elements. <i>Exception: </i> If it's a table row tag, then it is searched
     * for child nodes having a "dimdom" attribute. If not, it's a normal data
     * table row which is filled as long as there is more data filled in by the
     * programmer. If so, the current table row contains a microtemplate, i.e. a
     * "repetition within the repetition". The first case is easy: The child
     * nodes of type td or th are searched and counted, their text content is
     * removed and they're stored together with the number of found td/th child
     * nodes for their later/final removing. The second case is more complicated
     * because of the microtemplate substructure: The engine must then repeat
     * the repeated nodes within a table cell before it's allowed to jump to the
     * next cell.<br>
     * Note: The DimDomNodes are cleaned within this method, but the repeated
     * child nodes of a micro-template structure are cleaned within the
     * constructor of RepeatedElement.
     *
     * @param hIDNodeList  the list of nodes having a defined id attribute
     */
    private void clearTemplateElements( Hashtable hIDNodeList )
    {
        hRepeatedTableRows.clear();
        Enumeration eKeys = hIDNodeList.keys();
        while ( eKeys.hasMoreElements() )
        {
            String sNodeIdentifier = ( String )eKeys.nextElement();
            Element nextElement = ( Element )hIDNodeList.get( sNodeIdentifier );
            if ( isDimDomListNode( nextElement ) )
            {
                if ( !isTableRowNode( nextElement ) )
                    removeChildElements( nextElement );
                else
                {
                    //in case it's a table row, find out how many td or th tags are in there and
                    //store the node identifier together with this value in the Hashtable; then step
                    //through the td/th tags and remove each child text node
                    hRepeatedTableRows.put( sNodeIdentifier,
                            new Integer( getTableDataChildCount( nextElement ) ) );
                }
            }
        }
    }

    /**
     * Removes the child nodes that are Text nodess.
     *
     * @param node  the node to remove the child Texts of
     */
    protected void removeChildTextNodes( Node node )
    {
        for ( Node childOfChild = node.getFirstChild();
                childOfChild != null; childOfChild = childOfChild.getNextSibling() )
        {
            try
            {
                Text textToDelete = ( Text )childOfChild;
                textToDelete.getParentNode().removeChild( textToDelete );
            } catch ( Exception e )
            {
            }
        }
    }

    /**
     * Removes the child nodes that are Element nodess.
     *
     * @param node  the node to remove the child Elements of
     */
    protected void removeChildElements( Node node )
    {
        for ( Node childOfChild = node.getFirstChild();
                childOfChild != null; childOfChild = childOfChild.getNextSibling() )
        {
            try
            {
                Element elementToDelete = ( Element )childOfChild;
                elementToDelete.getParentNode().removeChild( elementToDelete );
            } catch ( Exception e )
            {
            }
        }
    }

    /**
     * (Re)Builds the internal list of micro-template nodes, i.e. those having a
     * child node with a set dimdom attribute. By DimDom specification, a
     * repeated tag <i>can't be</i> neither the html nor the head nor the body
     * nor a form tag.
     *
     * @param hClonedIDNodeList    list of id nodes to check
     * @exception DimDomException  thrown by the command parser
     */
    protected final void buildMicroTemplateNodeList( Hashtable hClonedIDNodeList )
        throws DimDomException
    {
        hMicroTemplateNodes.clear();

        Enumeration eIDs = hClonedIDNodeList.keys();
        while ( eIDs.hasMoreElements() )
        {
            String sNextIdentifier = ( String )eIDs.nextElement();
            Node nextNode = ( Node )hClonedIDNodeList.get( sNextIdentifier );
            if ( !sNextIdentifier.equalsIgnoreCase( "body" ) &&
                    !sNextIdentifier.equalsIgnoreCase( "head" ) &&
                    !sNextIdentifier.equalsIgnoreCase( "html" ) )
            {
                Vector vNextRE = new Vector();
                findAndAddRepeatedElements( nextNode.getChildNodes(), vNextRE );
                if ( !vNextRE.isEmpty() )
                    hMicroTemplateNodes.put( sNextIdentifier,
                            ( RepeatedElement[] )vNextRE.toArray( new RepeatedElement[]{} ) );
            }
        }
    }

	/**
	 * Returns whether the given test string is null or equals the empty string.
	 *
	 * @param sTestString   the String to be tested
	 * @return  whether the test string is empty or not
	 */
	protected final boolean isEmpty( String sTestString )
	{
		return ( sTestString == null || sTestString.equals( "" ) );
	}

//-------------------------------- inner class RepeatedElement ---------------------------------//
    /**
     * This simple class stored the template information on a repeated element,
     * i.e. a node that has a defined dimdom attribute, defining the maximum
     * number of repetitions that may be made on the same node.
     *
     * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
     *      Florian Steinsiepe</a>
     * @version   1.0
     */
    protected class RepeatedElement
    {
        /** the node having a dimdom attribute in the template */
        private Node repeatedNode;
        /** maximum number of allowed repetitions or -1, if unlimited */
        private int iMaxRepetitions;
        /** current number of processed repetitions on the repeatedNode */
        private int iNProcessedRepetitions;
        /** used in case of a repeated DimDomNode only */
        private int iTableDataNodeCount;

        /**
         * Stores the underlying, repeated node pointer and calls the
         * CommandParser to find out how many repetitions may be maximally made
         * on this node. In case the node is a DimDom-relevant node, it's
         * cleaned up first, e.g. the existing option tags from the template are
         * removed.
         *
         * @param repeatedNode         the node that contains a micro-template
         *      structure
         * @param sRepeatCommand       the value of the dimdom attribute in the
         *      template
         * @exception DimDomException  thrown by the command parser
         */
        protected RepeatedElement( Node repeatedNode, String sRepeatCommand )
            throws DimDomException
        {
            // store pointer and calculate the number of repetitions
            this.repeatedNode = repeatedNode;
            this.iMaxRepetitions = CommandParser.getNumberOfRepetitions( sRepeatCommand );

            // if it's a DimDomNode itself, clean up the child elements
            if ( isDimDomListNode( repeatedNode ) )
            {
                if ( !isTableRowNode( repeatedNode ) )
                    removeChildElements( repeatedNode );
                else
                {
                    //in case it's a table row, find out how many td or th tags are in there and
                    //store it; then step through the td/th tags and remove each child text node
                    NodeList childNodes = repeatedNode.getChildNodes();
                    for ( int i = 0; i < childNodes.getLength(); i++ )
                    {
                        Node nextChildNode = childNodes.item( i );
                        if ( isTableDataNode( nextChildNode ) )
                        {
                            //notice and...
                            this.iTableDataNodeCount++;

                            //...remove child text nodes
                            removeChildTextNodes( nextChildNode );
                        }
                    }
                }
            }
        }

        /**
         * Returns the number of repetitions that may be made on this node or
         * -1, if unlimited.
         *
         * @return   the value of the dimdom attribute's repeat command in the
         *      template
         */
        public int getMaxRepetitionCount()
        {
            return iMaxRepetitions;
        }

        /**
         * Returns the underlying, to-be-repeated node of the document that this
         * class was created for (may be either the original or the cloned
         * Document object!).
         *
         * @return   the underlying, to-be-repeated node
         */
        public Node getRepeatedNode()
        {
            return repeatedNode;
        }

        /**
         * Returns the (current) amount of repetition cycles that have been
         * processed on this class' repeated node until now. 0 means that the
         * class haven't been repeated yet, i.e. the method processRepetiton has
         * never been called.
         *
         * @return   the number of already processed repetition cycles
         */
        protected int getProcessedRepetitionsCount()
        {
            return iNProcessedRepetitions;
        }

        /**
         * Notifies this class that it's repeated node has been repeated / is
         * now in process of repetition. This leads to an update of the current
         * number of repetitions.
         */
        protected void processRepetition()
        {
            iNProcessedRepetitions++;
        }

        /**
         * Returns whether this class' repeated node is (still) allowed to be
         * repeated. This depends on the node's dimdom attribute in the
         * template: If the node is allowed to be repeated unlimitedly, this
         * method returns true anyway, else true is returned only if the number
         * of processed repetitions is lower than the maximum allowed number of
         * repetitions.
         *
         * @return   whether the node is (still) allowed to be repeated
         */
        protected boolean isRepeatable()
        {
            if ( ( this.getMaxRepetitionCount() == -1 ) ||
                    ( this.getProcessedRepetitionsCount() < this.getMaxRepetitionCount() ) )
                return true;
            else
                return false;
        }
    }
}
