package org.alltimeflashdreamer.dimdom;

import java.net.URL;
import java.util.*;

/**
 * Title: AbstractXHTMLSiteData<br>
 * Description: This is the superclass of each DOM processor using DimDom's own
 * processing mechanism. This means that it works together tightly with the
 * DimDomNodeImpl class which has a set of package protected methods, visible
 * for e.g. this class.
 *
 * @author    Copyright (c) 2001/02 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

public abstract class AbstractXHTMLSiteData implements XHTMLSiteData
{
    /** the name of the template file */
    private String sFileName;
    /** the URL of the template file */
    private URL url;
    /** the XHTML string */
    private String sXHTML;
    /** the ordered list of DimDomNodeImpl objects */
    protected Vector vDimDomNodeImpls = new Vector();

//-------------------------- XHTMLSiteData interface methods -------------------------------//
    /**
     * Returns a string that describes the XHTML source. This string must
     * identify the source, not the instance of this class! The extending class
     * must verify that one of the possible sources is set and not null, else a
     * NullPointerException will be thrown.
     *
     * @return   the class' identifier, to allow DimDom to recognize it again
     */
    public String getIdentifier()
    {
        if ( isFileSource() )
            return sFileName;
        else if ( isStringSource() )
            return sXHTML;
        else
            return url.toString();
    }

    /**
     * (is abstract to be implemented by the extending class)
     *
     * @return   the name of the available template processor class; for
     *      SimpleSiteData, this is "org.alltimeflashdreamer.dimdom.SimpleXHTMLDocument"
     */
    public abstract String getXHTMLDocumentClassName();

    /**
     * Returns the next implementing object of the DimDomNode interface, to
     * allow the DimDom engine to <i>process the next node</i> (or the next
     * clone of a micro-template node). Don't use this method if you want to
     * create a new DimDomNodeImpl, i.e. if you want to add a DimDom command
     * object to be processed by calling this method!
     *
     * @return   the next DimDomNodeImpl instance or null
     */
    public DimDomNode getNextDimDomNode()
    {
        if ( vDimDomNodeImpls.isEmpty() )
            return null;
        else
            return ( DimDomNodeImpl )vDimDomNodeImpls.remove( 0 );
    }

    /**
     * Returns the source string to be parsed.
     *
     * @return   the string set in setString or null
     */
    public String getString()
    {
        return sXHTML;
    }

    /**
     * Returns the file name of the template file to be parsed.
     *
     * @return   the file name set in setTemplateFilename or null
     */
    public String getTemplateFileName()
    {
        return sFileName;
    }

    /**
     * Returns the source URL to parse from.
     *
     * @return   the URL set in setURL or null
     */
    public URL getURL()
    {
        return url;
    }

//-------------------------- implementation specific methods -------------------------------//
    /**
     * Sets the internal template file name value and the others.
     *
     * @param sTemplateFilename  the name of the template file to be parsed
     */
    public void setTemplateFilename( String sTemplateFilename )
    {
        this.sFileName = sTemplateFilename;
        this.url = null;
        this.sXHTML = null;
    }

    /**
     * Sets the internal URL value and unsets the others.
     *
     * @param url  the URL of the URL to be parsed
     */
    public void setURL( URL url )
    {
        this.sFileName = null;
        this.url = url;
        this.sXHTML = null;
    }

    /**
     * Sets the internal string value and unsets the others.
     *
     * @param sInput  the string to be parsed
     */
    public void setString( String sInput )
    {
        this.sFileName = null;
        this.url = null;
        this.sXHTML = sInput;
    }

    /**
     * Returns whether the input source is (currently) a file or not.
     *
     * @return   is the template a file or not?
     */
    protected boolean isFileSource()
    {
        return ( sFileName != null );
    }

    /**
     * Returns whether the input source is (currently) just a string or not.
     *
     * @return   is the template a string or not?
     */
    protected boolean isStringSource()
    {
        return ( sXHTML != null );
    }

    /**
     * Returns whether the input source is (currently) an URL or not.
     *
     * @return   is the template an URL or not?
     */
    protected boolean isURLSource()
    {
        return ( url != null );
    }

    /**
     * Returns whether there are more DimDomNode implementing instances in the
     * pipeline or not.
     *
     * @return   are there more nodes to process?
     */
    protected boolean hasMoreDimDomNodeImpls()
    {
        return !vDimDomNodeImpls.isEmpty();
    }
}
