package org.alltimeflashdreamer.dimdom;

/**
 * Title: CommandParser<br>
 * Description: This class contains the methods that are able to parse DimDom
 * commands.<br>
 * These methods are centralized herein to easily find them again in case of
 * changes or extensions of the DimDom specification.
 *
 * @author    Copyright (c) 2001 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

class CommandParser
{
    private final static String ERROR1A = "Check of \"repeat\" command: Value in braces (";
    private final static String ERROR1E = ") is neither an int > 0 nor an \"*\"!";

    /** Constructor must not be called.  */
    private CommandParser() { }

    /**
     * Parses the command string and returns the number of repetitions to be
     * done / that are allowed at most.<br>
     * This method is defined statically to be callable from not instantiating
     * classes too and to increase the performance of this very simple task.<br>
     * In case the parsing fails, a DimDomException is thrown.
     *
     * @param sRepeatCommand    the value of the "dimdom" attribute in the
     *      template
     * @return                  the number of repetitions, -1 if unlimited ("*")
     * @throws DimDomException  thrown if the command isn't parseable
     */
    static int getNumberOfRepetitions( String sRepeatCommand )
        throws DimDomException
    {
        int iReturnValue = 0;

        if ( ( sRepeatCommand.startsWith( "repeat(" ) ) && ( sRepeatCommand.endsWith( ")" ) ) )
        {
            String sBetweenBrackets = sRepeatCommand.substring( "repeat(".length(), sRepeatCommand.length() - 1 );
            if ( sBetweenBrackets.equals( "*" ) )
                iReturnValue = -1;
            else
            {
                try
                {
                    iReturnValue = Integer.parseInt( sBetweenBrackets );
                    if ( ( iReturnValue <= 0 ) && ( iReturnValue != -1 ) )
                        throw new DimDomException( ERROR1A + sBetweenBrackets + ERROR1E );
                } catch ( NumberFormatException nfe )
                {
                    throw new DimDomException( ERROR1A + sBetweenBrackets + ERROR1E, nfe );
                }
            }
        }
        else
            throw new DimDomException( "Syntax error in " + sRepeatCommand +
                    "!\nUsage is \"repeat(* | n)\", with n as an int > 0." );

        return iReturnValue;
    }
}
