package org.alltimeflashdreamer.dimdom;

import java.io.*;
import java.util.*;
import org.xml.sax.SAXException;
import org.w3c.dom.Document;

/**
 * Title: DimDom<br>
 * Description: This is the "server core" of the DimDom engine. It works
 * statically somewhere in the JVM and collects instances of XHTMLDocument
 * implementing classes, each for one template source file or URL. If the print
 * method is called, DimDom tries to initialize the processing class and then to
 * call the processor method of this class. If the current instance is able to
 * do its job, the fully qualified XHTML string is returned, else the
 * appropriate exception is thrown.<br>
 * To allow XHTMLSiteData implementing classes to test the programmer's input,
 * DimDom provides the getTestDocument method which returns a fully qualified
 * DOM tree, wrapped in a org.w3c.dom.Document instance. This Document may then
 * be used to apply modifications of the DOM tree to find out whether the action
 * leads to a DOMException or is fully XHTML compliant.
 *
 * @author    Copyright (c) 2001/02 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

public final class DimDom
{
    /**
     * Stores a validateable XHTML String which is used to test the programmer's
     * commands. The validating DTD is the official one, provided by w3.org:<br>
	 * <a href="http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd</a>
     */
    final static String TESTDOCUMENT = "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n" +
            "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n" +
            "<html><head></head><body></body></html>";

    /**
     * the parsed test document pointer, ready to be cloned and returned on
     * access of the getTestDocument method
     */
    private static Document testDoc;

    /** the list of already loaded instances of XHTMLSiteData */
    private static Hashtable hLoadedSites = new Hashtable();

    /** Constructor must not be instantiated. */
    private DimDom() { }

    /**
     * Does everything that is necessary to be done to form a dynamically
     * updated XHTML output out of the values defined in SiteData.<br>
     * Therefore, the internal list of preloaded sites is checked against the
     * String received from the XHTMLSiteData's getIdentifier method. If such an
     * entry is found in the list, the object is received from the list and
     * reused, else a new instance of the class received from the
     * getXHTMLDocumentClass method is created.
     *
     * @param siteData             a class that implements
     *      org.alltimeflashdreamer.dimdom.XHTMLSiteData
     * @return                     the full XHTML string of the dynamically
     *      modified template
     * @exception SAXException     thrown by the parser
     * @exception IOException      thrown by the parser (input source)
     * @exception DimDomException  thrown by the processor, if something is/went
     *      wrong
     */
    public static String print( XHTMLSiteData siteData )
        throws SAXException, IOException, DimDomException
    {
        XHTMLDocument doc = ( XHTMLDocument )getDimDomDocument( siteData.getIdentifier() );

        //initialize and store a new implementing instance of XHTMLDocument, if necessary
        if ( doc == null )
        {
            try
            {
                doc = ( XHTMLDocument )Class.forName( siteData.getXHTMLDocumentClassName() ).newInstance();
            } catch ( Exception ex )
            {
                throw new DimDomException( "Couldn't create an instance of " +
                        siteData.getXHTMLDocumentClassName() + "!", ex );
            }

            addDimDomDocument( siteData.getIdentifier(), doc );
        }

        //the setter method initalizes the full DimDom processing mechanism
        doc.setSiteData( siteData );

        return doc.print();
    }

    /**
     * Returns a clone of the internal test document, ready to be used for e.g.
     * testing purposes
     *
     * @return   a fully XHTML compliant Document instance
     */
    public static Document getTestDocument()
    {
        if ( testDoc == null )
        {
            try
            {
                testDoc = new XHTMLParser().parse( TESTDOCUMENT );
            } catch ( Exception ex )
            {
                //MUST NOT HAPPEN, because TESTDOCUMENT is valid
                ex.printStackTrace();
            }
        }
        return ( Document )testDoc.cloneNode( true );
    }

    /**
     * Returns the associated object to sIdentifier from within the internal
     * list. If this list doesn't contain such a key, null is returned.
     *
     * @param sIdentifier  as received by the getIdentifer method of the
     *      XHTMLSiteData interface
     * @return             the required instance of XHTMLDocument
     */
    public static synchronized XHTMLDocument getDimDomDocument( String sIdentifier )
    {
        return ( XHTMLDocument )hLoadedSites.get( sIdentifier );
    }

    /**
     * Adds a key/value pair to the internal list. If this list already contains
     * such a key, the former value is overwritten.
     *
     * @param sIdentifier  as received by the getIdentifer method of the
     *      XHTMLSiteData interface
     * @param doc          the required instance of XHTMLDocument
     */
    public static synchronized void addDimDomDocument(
            String sIdentifier, XHTMLDocument doc )
    {
        hLoadedSites.put( sIdentifier, doc );
    }

    /**
     * Removes the associated object of sIdentifier from within the internal
     * list. If this list doesn't contain such a key, nothing is done.
     *
     * @param sIdentifier  as received by the getIdentifer method of the
     *      XHTMLSiteData interface
     */
    public static synchronized void removeDimDomDocument( String sIdentifier )
    {
        hLoadedSites.remove( sIdentifier );
    }
}
