package org.alltimeflashdreamer.dimdom;

import java.util.Vector;

/**
 * Title: DimDomNode<br>
 * Description: This interface collects the supported operations of the DimDom
 * engine that the programmer of the dynamic content is able to use.
 *
 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

public interface DimDomNode
{
    /**
     * Sets an attribute key/value pair to the node. If this attribute is
     * already defined for the underlying node, it is replaced during the
     * processing of the template. If the key or the value parameter is not
     * XHTML compliant, a DimDomException is thrown. If the boolean is true,
     * this attribute is added to each node of the same kind as the underlying
     * node.
     *
     * @param sKey                 the name of the attribute
     * @param sValue               the value of the attribute
     * @param bIsCommon            whether to set this attribute to each node
     *      having the same node name as the template node having this class'
     *      node identifier
     * @exception DimDomException  if one of the parameters is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setAttribute( String sKey, String sValue, boolean bIsCommon )
        throws DimDomException;

    /**
     * Sets an attribute key/value pair to the node. If this attribute is
     * already defined for the underlying node, it is replaced during the
     * processing of the template. If the key or the value parameter is not
     * XHTML compliant, a DimDomException is thrown. This method is designed for
     * individual nodes only.
     *
     * @param sKey                 the name of the attribute
     * @param sValue               the value of the attribute
     * @exception DimDomException  if one of the parameters is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setAttribute( String sKey, String sValue )
        throws DimDomException;

    /**
     * Sets an attribute key/value pair to DimDom-specific child of node. If
     * this attribute is already defined for the child node, it is replaced
     * during the processing of the template. If the key or the value parameter
     * is not XHTML compliant, a DimDomException is thrown. This method is
     * designed for individual nodes only, because DimDom doesn't append
     * children to common nodes.
     *
     * @param sKey                 the name of the attribute
     * @param sValue               the value of the attribute
     * @exception DimDomException  if one of the parameters is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setChildAttribute( String sKey, String sValue )
        throws DimDomException;

    /**
     * Stores a single String value, without selection.
     *
     * @param sValue  just a value
     */
    public void addData( String sValue );

    /**
     * Stores a single String value, with selection.
     *
     * @param sValue       just a value
     * @param bIsSelected  whether to select the value or not
     */
    public void addData( String sValue, boolean bIsSelected );

    /**
     * Stores a list of String values, without selection.
     *
     * @param sValues  a value list
     */
    public void addData( String[] sValues );

    /**
     * Stores a list of String values, without selection.
     *
     * @param vValues  a value list
     */
    public void addData( Vector vValues );

    /**
     * Stores a list of String values, with selection index.
     *
     * @param sValues         a value list
     * @param iSelectedIndex  an existing index, else nothing is selected
     */
    public void addData( String[] sValues, int iSelectedIndex );

    /**
     * Stores a list of String values, with selected string.
     *
     * @param sValues         a value list
     * @param sSelectedEntry  an existing string of the list, else nothing is
     *      selected
     */
    public void addData( String[] sValues, String sSelectedEntry );

    /**
     * Stores a list of <i>String</i> values, with selection index.
     *
     * @param vValues         a value list
     * @param iSelectedIndex  an existing index, else nothing is selected
     */
    public void addData( Vector vValues, int iSelectedIndex );

    /**
     * Stores a list of <i>String</i> values, with selected string.
     *
     * @param vValues         a value list
     * @param sSelectedEntry  an existing string of the list, else nothing is
     *      selected
     */
    public void addData( Vector vValues, String sSelectedEntry );

    /**
     * Stores a field (i.e. a 2-dimensional list) of String values, without
     * selection.
     *
     * @param sValues2D  a value field
     */
    public void addData( String[][] sValues2D );

    /**
     * Stores a field (i.e. a 2-dimensional list) of String values, with
     * selected row.
     *
     * @param sValues2D     a value field
     * @param iSelectedRow  each value of this row (starting with 0) will be
     *      selected in the output
     */
    public void addData( String[][] sValues2D, int iSelectedRow );

    /**
     * Stores a field (i.e. a 2-dimensional list) of String values, with
     * selected row and column.
     *
     * @param sValues2D        a value field
     * @param iSelectedRow     the row index (starting with 0) of the selected
     *      element
     * @param iSelectedColumn  the column index (starting with 0) of the
     *      selected element
     */
    public void addData( String[][] sValues2D, int iSelectedRow, int iSelectedColumn );

    /**
     * Sets the visibility of the DimDomNode. If hidden, its underlying node is
     * removed from the DOM tree during processing to make it invisible.
     *
     * @param bIsVisible  true for visible, false for invisible
     */
    public void setVisible( boolean bIsVisible );

    /**
     * Sets the visibility of the DimDomNode to false and updates the
     * replacement value. If the node is still hidden during processing, its
     * underlying node is replaced with a Text node having the given replacement
     * as text.
     *
     * @param sReplacingText  e.g. a simple "-" to demonstrate there is no more
     *      of what has been there before
     */
    public void setReplacingText( String sReplacingText );

    /**
     * Embeds the underlying DOM node with a "a" element and sets the parameter
     * String to the element's "href" attribute. The given string may of course
     * be a JavaScript command or something else.
     *
     * @param sTargetURL           the URL string to set to the link's href
     *      attribute
     * @exception DimDomException  if the target URL string is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setEmbeddingLink( String sTargetURL )
        throws DimDomException;

    /**
     * Embeds the underlying DOM node with a "a" element and sets the parameter
     * String to the element's "href" attribute. The given string may of course
     * be a JavaScript command or something else.
     *
     * @param sTargetURL           the URL string to set to the link's href
     *      attribute
     * @param sTargetFrame         the browser frame name where the link shall
     *      point to
     * @exception DimDomException  if the target URL string is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setEmbeddingLink( String sTargetURL, String sTargetFrame )
        throws DimDomException;
}
