package org.alltimeflashdreamer.dimdom;

import java.util.*;
import org.w3c.dom.*;

/**
 * Title: DimDomNodeImpl<br>
 * Description: This class represents one node that is processed by the DimDom
 * engine. Its constructor stores the identifier which <i>must exist</i> in the
 * template file. If the template lacks this String (case-insensitive
 * comparison), a DimDomException will be thrown on accessing that node.
 *
 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

public class DimDomNodeImpl implements DimDomNode
{
    /** unique identifier, must exist in the template too (case-insensitive) */
    private String sNodeIdentifier;
    /** the Document to run tests with */
    private Document testDoc;
    /** list of key/value pairs that represent node attributes */
    private Vector vAttributes = new Vector();
    /** the ordered list of data objects; each entry is an instance of ValueObject */
    private Vector vValueObjects = new Vector();
    /**
     * the replacement string; if null, the node is visible; if "", the
     * setInvisible() method was called, else the setInvisible(String) method
     */
    private String sReplacement = null;
    /** the embedding link, as an instance of EmbeddingLink */
    private EmbeddingLink embeddingLink = null;

    /**
     * @param sNodeIdentifier  as defined in the template
     * @param testDocument     a valid XHTML Document object which is used to
     *      run tests while setting data/attributes
     */
    public DimDomNodeImpl( Document testDocument, String sNodeIdentifier )
    {
        this.sNodeIdentifier = sNodeIdentifier.toLowerCase();
        this.testDoc = testDocument;
    }

//------------------------------ DimDomNode interface methods -----------------------------------//
    /**
     * Sets an attribute key/value pair to the node. If this attribute is
     * already defined for the underlying node, it is replaced during the
     * processing of the template. If the key or the value parameter is not
     * XHTML compliant, a DimDomException is thrown If the boolean is true, this
     * attribute is added to each node of the same kind as the underlying node.
     *
     * @param sKey                 the name of the attribute
     * @param sValue               the value of the attribute
     * @param bIsCommon            whether to set this attribute to each node
     *      having the same node name as the template node having this class'
     *      node identifier
     * @exception DimDomException  if one of the parameters is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setAttribute( String sKey, String sValue, boolean bIsCommon )
        throws DimDomException
    {
        setAttribute( sKey, sValue, bIsCommon, false );
    }

    /**
     * Sets an attribute key/value pair to the node. If this attribute is
     * already defined for the underlying node, it is replaced during the
     * processing of the template. If the key or the value parameter is not
     * XHTML compliant, a DimDomException is thrown. This method is designed for
     * individual nodes only.
     *
     * @param sKey                 the name of the attribute
     * @param sValue               the value of the attribute
     * @exception DimDomException  if one of the parameters is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setAttribute( String sKey, String sValue )
        throws DimDomException
    {
        setAttribute( sKey, sValue, false, false );
    }

    /**
     * Sets an attribute key/value pair to DimDom-specific child of node. If
     * this attribute is already defined for the child node, it is replaced
     * during the processing of the template. If the key or the value parameter
     * is not XHTML compliant, a DimDomException is thrown. This method is
     * designed for individual nodes only, because DimDom doesn't append
     * children to common nodes.
     *
     * @param sKey                 the name of the attribute
     * @param sValue               the value of the attribute
     * @exception DimDomException  if one of the parameters is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setChildAttribute( String sKey, String sValue )
        throws DimDomException
    {
        setAttribute( sKey, sValue, false, true );
    }

    /**
     * Stores a single String value, without selection.
     *
     * @param sValue  just a value
     */
    public void addData( String sValue )
    {
        vValueObjects.add( new ValueObject( sValue ) );
    }

    /**
     * Stores a single String value, with selection.
     *
     * @param sValue       just a value
     * @param bIsSelected  whether to select the value or not
     */
    public void addData( String sValue, boolean bIsSelected )
    {
        vValueObjects.add( new ValueObject( sValue, bIsSelected ) );
    }

    /**
     * Stores a list of String values, without selection.
     *
     * @param sValues  a value list
     */
    public void addData( String[] sValues )
    {
        vValueObjects.add( new ValueObject( sValues ) );
    }

    /**
     * Stores a list of String values, without selection.
     *
     * @param vValues  a value list
     */
    public void addData( Vector vValues )
    {
        vValueObjects.add( new ValueObject( vValues ) );
    }

    /**
     * Stores a list of String values, with selection.
     *
     * @param sValues         a value list
     * @param iSelectedIndex  an existing index, else nothing is selected
     */
    public void addData( String[] sValues, int iSelectedIndex )
    {
        vValueObjects.add( new ValueObject( sValues, iSelectedIndex ) );
    }

    /**
     * Stores a list of String values, with selected string.
     *
     * @param sValues         a value list
     * @param sSelectedEntry  an existing string of the list, else nothing is
     *      selected
     */
    public void addData( String[] sValues, String sSelectedEntry )
    {
        vValueObjects.add( new ValueObject( sValues, getIndex( sValues, sSelectedEntry ) ) );
    }

    /**
     * Stores a list of String values, without selection.
     *
     * @param vValues         a value list
     * @param iSelectedIndex  an existing index, else nothing is selected
     */
    public void addData( Vector vValues, int iSelectedIndex )
    {
        vValueObjects.add( new ValueObject( vValues, iSelectedIndex ) );
    }

    /**
     * Stores a list of <i>String</i> values, with selected string.
     *
     * @param vValues         a value list
     * @param sSelectedEntry  an existing string of the list, else nothing is
     *      selected
     */
    public void addData( Vector vValues, String sSelectedEntry )
    {
        vValueObjects.add( new ValueObject( vValues,
                getIndex( ( String[] )vValues.toArray( new String[]{} ), sSelectedEntry ) ) );
    }

    /**
     * Stores a field (i.e. a 2-dimensional list) of String values, without
     * selection.
     *
     * @param sValues2D  a value field
     */
    public void addData( String[][] sValues2D )
    {
        vValueObjects.add( new ValueObject( sValues2D ) );
    }

    /**
     * Stores a field (i.e. a 2-dimensional list) of String values, with
     * selected row.
     *
     * @param sValues2D     a value field
     * @param iSelectedRow  each value of this row (starting with 0) will be
     *      selected in the output
     */
    public void addData( String[][] sValues2D, int iSelectedRow )
    {
        vValueObjects.add( new ValueObject( sValues2D, iSelectedRow ) );
    }

    /**
     * Stores a field (i.e. a 2-dimensional list) of String values, with
     * selected row and column.
     *
     * @param sValues2D        a value field
     * @param iSelectedRow     the row index (starting with 0) of the selected
     *      element
     * @param iSelectedColumn  the column index (starting with 0) of the
     *      selected element
     */
    public void addData( String[][] sValues2D, int iSelectedRow, int iSelectedColumn )
    {
        vValueObjects.add( new ValueObject( sValues2D, iSelectedRow, iSelectedColumn ) );
    }

    /**
     * Sets the visibility of the DimDomNode. If hidden, its underlying node is
     * removed from the DOM tree during processing to make it invisible.
     *
     * @param bIsVisible  true for visible, false for invisible
     */
    public void setVisible( boolean bIsVisible )
    {
        if ( bIsVisible )
            sReplacement = null;
        else
            sReplacement = "";
    }

    /**
     * Sets the visibility of the DimDomNode to false and updates the
     * replacement value. If the node is still hidden during processing, its
     * underlying node is replaced with a Text node having the given replacement
     * as text.
     *
     * @param sReplacingText  e.g. a simple "-" to demonstrate there is no more
     *      of what has been there before
     */
    public void setReplacingText( String sReplacingText )
    {
        this.sReplacement = sReplacingText;
    }

    /**
     * Embeds the underlying DOM node with a "a" element and sets the parameter
     * String to the element's "href" attribute. The given string may of course
     * be a JavaScript command or something else.
     *
     * @param sTargetURL           the URL string to set to the link's href
     *      attribute
     * @exception DimDomException  if the target URL string is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setEmbeddingLink( String sTargetURL )
        throws DimDomException
    {
        setEmbeddingLink( sTargetURL, null );
    }

    /**
     * Embeds the underlying DOM node with a "a" element and sets the parameter
     * String to the element's "href" attribute. The given string may of course
     * be a JavaScript command or something else.
     *
     * @param sTargetURL           the URL string to set to the link's href
     *      attribute
     * @param sTargetFrame         the browser frame name where the link shall
     *      point to
     * @exception DimDomException  if the target URL string is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    public void setEmbeddingLink( String sTargetURL, String sTargetFrame )
        throws DimDomException
    {
        Element testLink = null;
        Attr href = null;
        Attr target = null;
        try
        {
            testLink = testDoc.createElement( "a" );
            href = testDoc.createAttribute( "href" );
            href.setNodeValue( sTargetURL );
            testLink.setAttributeNode( href );
        } catch ( DOMException de )
        {
            throw new DimDomException( "Target string " + sTargetURL + " is not XHTML compliant!", de );
        }
        if ( sTargetFrame != null )
        {
            try
            {
                target = testDoc.createAttribute( "target" );
                target.setNodeValue( sTargetFrame );
                testLink.setAttributeNode( target );
            } catch ( DOMException de )
            {
                throw new DimDomException( "Target frame " + sTargetFrame + " is not XHTML compliant!", de );
            }
        }
        if ( testLink == null )
            //worst case scenario...
            throw new DimDomException( "Couldn't create link node! Please check your JVM and DimDom installations." );
        embeddingLink = new EmbeddingLink( href, target );
    }

//------------------------------ implementation specific methods -----------------------------------//
    /**
     * Returns whether there are more instances of ValueObject in the pipeline
     * or not.
     *
     * @return   more ValueObject instances?
     */
    boolean hasMoreValueObjects()
    {
        return !vValueObjects.isEmpty();
    }

    /**
     * Returns an instance of ValueObject, to allow the DimDom engine to process
     * the next cycle on this class' underlying element/node.
     *
     * @return   the next stored instance of ValueObject or null
     */
    ValueObject getNextValueObject()
    {
        if ( !hasMoreValueObjects() )
            return null;
        else
            return ( ValueObject )vValueObjects.remove( 0 );
    }

    /**
     * Returns whether there are more instances of Attr in the pipeline or not.
     *
     * @return   more AttributeObject instances?
     */
    boolean hasMoreAttributeObjects()
    {
        return !vAttributes.isEmpty();
    }

    /**
     * Returns an AttributeObject instance to be appended to the underlying node
     * by the DimDom engine.
     *
     * @return   the next stored Attr or null
     */
    AttributeObject getNextAttributeObject()
    {
        if ( !hasMoreAttributeObjects() )
            return null;
        else
            return ( AttributeObject )vAttributes.remove( 0 );
    }

    /**
     * Returns the visibility of the node.
     *
     * @return   true if visible, else false
     */
    boolean isVisible()
    {
        return ( sReplacement == null );
    }

    /**
     * Returns the replacing text of this node. The returned value is null if
     * the node is set visible, "" if no replacement was defined and else the
     * replacing text.
     *
     * @return   null, "" or a text string
     */
    String getReplacingText()
    {
        return sReplacement;
    }

    /**
     * Returns whether this node was set to be embedded in a link.
     *
     * @return   is EmbeddedLink object defined?
     */
    boolean hasEmbeddingLink()
    {
        return ( this.embeddingLink != null );
    }

    /**
     * Returns this class EmbeddingLink instance, if defined
     *
     * @return   the EmbeddingLink instance or null
     */
    EmbeddingLink getEmbeddingLink()
    {
        return embeddingLink;
    }

    /**
     * Returns the value of the id attribute of the underlying (DimDom) node.
     *
     * @return   the identifying string, which is unique in the template
     */
    String getNodeIdentifier()
    {
        return sNodeIdentifier;
    }

    /**
     * (general implementation of the different attribute setter methods)
     *
     * @param sKey                 the name of the attribute
     * @param sValue               the value of the attribute
     * @param bIsCommon            whether to set this attribute to each node
     *      having the same node name as the template node having this class'
     *      node identifier
     * @param bIsChild             whether this attribute is to be appended ot a
     *      child node or not
     * @exception DimDomException  if one of the parameters is not XHTML
     *      compliant (see org.w3c.dom.Attr class)
     */
    private void setAttribute( String sKey, String sValue,
            boolean bIsCommon, boolean bIsChild )
        throws DimDomException
    {
        Attr newAttr = null;
        try
        {
            newAttr = testDoc.createAttribute( sKey );
        } catch ( DOMException de )
        {
            throw new DimDomException( "Key " + sKey + " is not XHTML compliant!", de );
        }
        try
        {
            newAttr.setNodeValue( sValue );
        } catch ( DOMException de )
        {
            throw new DimDomException( "Value " + sValue + " is not XHTML compliant!", de );
        }
        if ( newAttr == null )
            //must never happen, would be quite critical...
            throw new DimDomException( "Couldn't create attribute node! Please check your JVM and DimDom installations." );
        vAttributes.add( new AttributeObject( newAttr, bIsCommon, bIsChild ) );
    }

    /**
     * Searches the selected value within the list of values. If this value is
     * found (null values are not compared!), its index within the list is
     * returned, else -1, to indicate that the value doesn't occur in the list.
     *
     * @param sValues         a list of String values
     * @param sSelectedValue  the String to search for
     * @return                the index of the string within the list or -1
     */
    private int getIndex( String[] sValues, String sSelectedValue )
    {
        if ( ( sValues == null ) || ( sSelectedValue == null ) )
            return -1;
        for ( int i = 0; i < sValues.length; i++ )
        {
            if ( sValues[i].equals( sSelectedValue ) )
                return i;
        }
        return -1;
    }

//---------------------------- inner class ValueObject ---------------------------//
    /**
     * The ValueObject class is abled to store data in the supported formats of
     * DimDom, i.e. a simple String, a String[], a String[] with selection
     * index, a String[][], a String[][] with selection row and a String[][]
     * with selection index (row/col). <br>
     * Note: Each instance may store <i>only one</i> of the possibilities listed
     * above!
     *
     * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
     *      Florian Steinsiepe</a>
     * @version   1.0
     */
    class ValueObject
    {
        private int iSelectedRow = -1;
        private int iSelectedColumn = -1;
        private int i2DCtr = 0;
        private Vector vData;
        private String[][] sData2D;

        /**
         * Stores a single String value, without selection.
         *
         * @param sValue  just a value
         */
        private ValueObject( String sValue )
        {
            vData = new Vector();
            vData.add( sValue );
        }

        /**
         * Stores a single String value, with selection.
         *
         * @param sValue       just a value
         * @param bIsSelected  whether to select the entry or not
         */
        private ValueObject( String sValue, boolean bIsSelected )
        {
            this( sValue );
            if ( bIsSelected )
                this.iSelectedRow = 0;
        }

        /**
         * Stores a list of String values, without selection.
         *
         * @param sValues  a value list
         */
        private ValueObject( String[] sValues )
        {
            vData = new Vector();
            for ( int i = 0; i < sValues.length; i++ )
                vData.add( sValues[i] );
        }

        /**
         * Stores a list of String values, without selection.
         *
         * @param vValues  a value list
         */
        private ValueObject( Vector vValues )
        {
            this.vData = vValues;
        }

        /**
         * Stores a list of String values, with selection.
         *
         * @param sValues         a value list
         * @param iSelectedIndex  an existing index, else nothing is selected
         */
        private ValueObject( String[] sValues, int iSelectedIndex )
        {
            this( sValues );
            this.iSelectedRow = iSelectedIndex;
        }

        /**
         * Stores a list of String values, without selection.
         *
         * @param vValues         a value list
         * @param iSelectedIndex  an existing index, else nothing is selected
         */
        private ValueObject( Vector vValues, int iSelectedIndex )
        {
            this( vValues );
            this.iSelectedRow = iSelectedIndex;
        }

        /**
         * Stores a field (i.e. a 2-dimensional list) of String values, without
         * selection.
         *
         * @param sValues2D  a value field
         */
        private ValueObject( String[][] sValues2D )
        {
            this.sData2D = sValues2D;
        }

        /**
         * Stores a field (i.e. a 2-dimensional list) of String values, with
         * selected row.
         *
         * @param sValues2D     a value field
         * @param iSelectedRow  each value of this row (starting with 0) will be
         *      selected in the output
         */
        private ValueObject( String[][] sValues2D, int iSelectedRow )
        {
            this( sValues2D );
            this.iSelectedRow = iSelectedRow;
        }

        /**
         * Stores a field (i.e. a 2-dimensional list) of String values, with
         * selected row and column.
         *
         * @param sValues2D        a value field
         * @param iSelectedRow     the row index (starting with 0) of the
         *      selected element
         * @param iSelectedColumn  the column index (starting with 0) of the
         *      selected element
         */
        private ValueObject( String[][] sValues2D, int iSelectedRow, int iSelectedColumn )
        {
            this( sValues2D, iSelectedRow );
            this.iSelectedColumn = iSelectedColumn;
        }

        /**
         * Returns whether a value was selected or not.
         *
         * @return   is a data entry selected?
         */
        boolean hasSelection()
        {
            return ( ( this.iSelectedRow >= 0 ) || ( this.iSelectedColumn >= 0 ) );
        }

        /**
         * Returns whether the current element receives a single value or not.
         *
         * @return   one single value?
         */
        boolean isSingleValue()
        {
            if ( this.vData == null )
                return false;
            else
                return ( this.vData.size() == 1 );
        }

        /**
         * Returns whether the current element receives a one-dimensional list
         * of values or not.
         *
         * @return   a value list?
         */
        boolean isValueList()
        {
            if ( this.vData == null )
                return false;
            else
                return ( this.vData.size() > 1 );
        }

        /**
         * Returns whether the current element receives a two-dimensional list
         * of values or not.
         *
         * @return   a value field?
         */
        boolean isValueField()
        {
            return ( this.sData2D != null );
        }

        /**
         * Returns the index of the selected row, if a selection in a 1D or 2D
         * value list/field is set, else -1.
         *
         * @return   a valid index or -1
         */
        int getSelectedRow()
        {
            return this.iSelectedRow;
        }

        /**
         * Returns the index of the selected column, if a selection in a 2D
         * value field is set, else -1.
         *
         * @return   a valid index or -1
         */
        int getSelectedColumn()
        {
            return this.iSelectedColumn;
        }

        /**
         * Returns whether there are more values to add to the underlying node.
         *
         * @return   are there more values?
         */
        boolean hasMoreValues()
        {
            if ( vData != null )
                return ( !vData.isEmpty() );
            else if ( sData2D != null )
                return ( i2DCtr < ( sData2D.length * sData2D[0].length ) );
            else
                return false;
        }

        /**
         * Returns the next value to add to the underlying node or null, if no
         * more values are stored.
         *
         * @return   next value (as String) or null
         */
        String getNextValue()
        {
            if ( !hasMoreValues() )
                return null;
            else if ( vData != null )
                return ( String )vData.remove( 0 );
            else if ( sData2D != null )
            {
                int iRow = i2DCtr / sData2D[0].length;
                int iColumn = i2DCtr - ( iRow * sData2D[0].length );
                i2DCtr++;
                return sData2D[iRow][iColumn];
            }
            else
                return null;
        }

        /**
         * Returns the next value row to add to the underlying node or null, if
         * no more values are stored.
         *
         * @return   next value list (as String[]) or null
         */
        String[] getNextValueRow()
        {
            if ( !hasMoreValues() )
                return null;
            else if ( sData2D != null )
            {
                int iRow = i2DCtr / sData2D[0].length;
                i2DCtr += sData2D[0].length;
                try
                {
                    return sData2D[iRow];
                } catch ( ArrayIndexOutOfBoundsException aioobe )
                {
                    return null;
                }
            }
            else
                return null;
        }
    }

//----------------------------------- inner class EmbeddingLink ---------------------------------//
    /**
     * Simple object to stored the set values of an embedding link ("a"
     * element).
     *
     * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
     *      Florian Steinsiepe</a>
     * @version   1.0
     */
    class EmbeddingLink
    {
        private Attr href;
        private Attr target;

        /**
         * Sets the target URL and frame values.
         *
         * @param href    the Attr node named href (!= null)
         * @param target  the Attr node named target (may be null)
         */
        private EmbeddingLink( Attr href, Attr target )
        {
            this.href = href;
            this.target = target;
        }

        /**
         * Returns the target URL attribute node.
         *
         * @return   the Attr instance named href (is never null)
         */
        Attr getHref()
        {
            return this.href;
        }

        /**
         * Returns the target frame attribute node.
         *
         * @return   the Attr instance named target or null
         */
        Attr getTarget()
        {
            return this.target;
        }
    }

//----------------------------------- inner class AttributeObject ---------------------------------//
    /**
     * Description of the Class
     *
     * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
     *      Florian Steinsiepe</a>
     * @version   1.0
     */
    class AttributeObject
    {
        private Attr attr;
        private boolean bIsCommon;
        private boolean bIsChildAttribute;

        /**
         * Stores the two parameters.
         *
         * @param attr               the attribute node which is ready to be
         *      added
         * @param bIsCommon          whether this is a common attribute or not
         * @param bIsChildAttribute  whether to set this attribute to each node
         *      having the same node name as the template node having this
         *      class' node identifier
         */
        private AttributeObject( Attr attr, boolean bIsCommon, boolean bIsChildAttribute )
        {
            this.attr = attr;
            this.bIsCommon = bIsCommon;
            this.bIsChildAttribute = bIsChildAttribute;
        }

        /**
         * Returns the Attr node to add to the underlying element.
         *
         * @return   the Attr node that is already ready to be added to the
         *      underlying node
         */
        Attr getAttr()
        {
            return this.attr;
        }

        /**
         * Returns whether this is a common attribute, which is set to each node
         * of the same type as the node having the node identifier of the parent
         * class, or not.
         *
         * @return   is this a common attribute?
         */
        boolean isCommon()
        {
            return this.bIsCommon;
        }

        /**
         * Returns whether this is a child attribute, which is set to
         * DimDom-specific child node, or not.
         *
         * @return   is this a child attribute?
         */
        boolean isChildAttribute()
        {
            return this.bIsChildAttribute;
        }
    }
}
