package org.alltimeflashdreamer.dimdom;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;
import org.xml.sax.*;

/**
 * Title: SimpleDocument<br>
 * Description: This class implements the "simple" logic of the DimDom,
 * extending the very "base" logic from AbstractHTMLDocument, which means that
 * it implements the abstract method getNextRepeatedElement method defined in
 * the superclass. This method is called in case the programmer coded more than
 * one DimDom command on a node having children with a set "dimdom" attribute,
 * telling the DimDom engine that they are ready to be multiplied/cloned on each
 * repeated command (i.e. on the same original template node).
 *
 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

public class SimpleDocument extends AbstractXHTMLDocument
{
    /**
     * Constructor is package protected and proceeds the full job list to be
     * immediately printable.
     *
     * @param siteData             the XHTMLSiteData implementing class
     * @exception SAXException     thrown by the parser
     * @exception IOException      thrown by the file system
     * @exception DimDomException  thrown in case something in the template
     *      isn't DimDom compliant
     */
    public SimpleDocument( XHTMLSiteData siteData )
        throws SAXException, IOException, DimDomException
    {
        this.setSiteData( siteData );
    }

    /** Constructor is empty and package protected. */
    SimpleDocument() { }

    /**
     * This method is the core of the SimpleDocument's "intelligence". Because
     * it is only called in case the designer added a "dimdom=repeat(n)"
     * attribute to at least one subnode of a node having an "id" attribute, its
     * "calling context" is very special in general. Therefore, the method first
     * checks the context and then reacts according to the situation, along the
     * rules listed below.<br>
     * The returned element is the one that is in turn to be filled according to
     * the "dimdom=repeat(n)" setting(s) in the template. Returned values are:
     *
     * <ul>
     *   <li> If the list of RepeatedElements is empty, null is returned
     *   <li> If the internal list of RepeatedElements contains only one element
     *   related to (i.e. a child node of the given element) element and the
     *   repeat mode is * (= -1), a clone of this single repeated element is
     *   returned
     *   <li> If the internal list of RepeatedElements contains multiple
     *   elements, the returned node is evaluated based on the following rules:
     *
     *   <ul>
     *     <li> Whether the current number of repetition on the same node,
     *     compared to the maximum allowed repetition value defined in the
     *     template (node), is still smaller or equal; if not, the next repeated
     *     element is asked, and so on
     *     <li> Whether it's a tr tag or not
     *   </ul>
     *
     * </ul>
     * Each child node is repeated "rowwise", i.e. the engine assumes that the
     * designer implemented a "micro-template" structure that shall be repeated
     * as a whole bunch. In case there is a microtemplate <i>within</i> one of
     * the tr element's subnodes (imagine one wants to dynamically change
     * multiple img nodes within a td tag), the engine rowwise adapts the
     * microtemplate's nodes and then clones the whole row, restarting with the
     * first repeated subnode of the new tr element. <br>
     * The new element, i.e. the current clone of a template node, is appended
     * to the template node (which will later be deleted, during printing of the
     * DOM) immediately.<br>
     * Important: If the cloned element is an input element (select, input or
     * button), it's asked for its name attribute. If this attribute is defined,
     * its value is modified by the current number of repetitions, <i>but only
     * </i> if this number is greater than 0.
     *
     * @param sNodeIdentifier      the id attribute value of the node to be
     *      processed
     * @param element              the current node to be processed, for
     *      deciding how to behave
     * @return                     the next repeated element, i.e. a clone that
     *      is already appended/inserted correctly
     * @exception DimDomException  wrapper for unexpected errors
     */
    protected Element getNextRepeatedElement( String sNodeIdentifier, Element element )
        throws DimDomException
    {
        Element returnValue = null;

        //the repetition counter indicates how many times this method has been called
        //on the same node, to allow this method to find out which element shall
        //be taken out of the list of repeated elements that are stored for the
        //given element; the value 0 indicates the first run on that node
        int iRepetitionCtr = this.getProcessedRepetitionsCount( sNodeIdentifier );

        //get list of those repeated elements that are related to element, return null
        //if no such elements are stored in the template
        RepeatedElement[] allRepeatedElements = this.getRepeatedElements( sNodeIdentifier );
        if ( allRepeatedElements == null )
            return returnValue;

        //run the cycle "rowwise", i.e. once for each member of the microtemplate per cycle
        int iCtr = 0;
        int iIndex = getIndex( iRepetitionCtr, allRepeatedElements );

        //in case it's a "microtemplate within a (child node of a) tr tag" situation
        //AND the cycle is at the beginning (iIndex == 0), clone the whole tr tag,
        //rebuild the allRepeatedElements list and set it as new element-related list
        //(little hack to mimic that everything is still as before)
        if ( ( iIndex == 0 ) && ( element.getNodeName().equalsIgnoreCase( "tr" ) ||
                ( this.isTableDataNode( element ) && containsTableNode( element ) ) ) )
        {
            Element clonedTR = ( Element )element.cloneNode( true );
            clonedTR.removeAttribute( "id" );
            element.getParentNode().appendChild( clonedTR );
            Vector vRepElems = new Vector();
            this.findAndAddRepeatedElements( clonedTR.getChildNodes(), vRepElems );
            allRepeatedElements = ( ( vRepElems.size() > 0 ) ?
                    ( RepeatedElement[] )vRepElems.toArray( new RepeatedElement[vRepElems.size()] ) : null );

            //the exception is thrown as a safety move; must never happen (not even come to here)!
            if ( allRepeatedElements == null )
                throw new DimDomException( "Cloned micro-template table row doesn't contain " +
                        "repeated elements! Please check your code and/or template." );
            this.setRepeatedElements( sNodeIdentifier, allRepeatedElements );
        }

        //search list of repeated elements from the beginning index on for the next
        //element that still is allowed be cloned and return this element; to avoid
        //the situation where a still repeatable node is "forgotten", because the
        //index is already to high, the index is reset (only) once
        boolean bIsFirstSearch = true;
        while ( true )
        {
            RepeatedElement currentRepeatedElement = allRepeatedElements[iIndex];
            if ( currentRepeatedElement.isRepeatable() )
            {
                currentRepeatedElement.processRepetition();
                returnValue = ( Element )currentRepeatedElement.getRepeatedNode().cloneNode( true );
                returnValue.removeAttribute( "dimdom" );
                checkNameAttribute( returnValue, currentRepeatedElement.getProcessedRepetitionsCount() );
                currentRepeatedElement.getRepeatedNode().getParentNode().appendChild( returnValue );
                break;
            }
            else
            {
                iIndex++;
                if ( iIndex >= allRepeatedElements.length )
                {
                    if ( bIsFirstSearch )
                    {
                        iIndex = 0;
                        bIsFirstSearch = false;
                    }
                    else
                        break;
                }
            }
        }

        //return the node, if found, else null
        return returnValue;
    }

    /**
     * Checks whether a node having a defined name attribute is repeated or not.
     * If so and if the name attribute value is not null or empty and if the
     * given int is > 0 and if the given element is an input tag, the repetition
     * counter is added to the name, to produce different URL parameter names,
     * for later refinding them in the correct order.
     *
     * @param element         the element to check
     * @param iRepetitionCtr  the current number or repetition
     */
    private void checkNameAttribute( Element element, int iRepetitionCtr )
    {
        if ( iRepetitionCtr == 0 )
            return;

        NamedNodeMap attrs = element.getAttributes();
        if ( ( attrs != null ) && ( isInputElement( element ) ) )
        {
            for ( int i = 0; i < attrs.getLength(); i++ )
            {
                if ( attrs.item( i ).getNodeName().equalsIgnoreCase( "name" ) )
                {
                    String sName = attrs.item( i ).getNodeValue();
                    if ( ( sName != null ) && ( !sName.equals( "" ) ) )
                        element.setAttribute( "name", sName + iRepetitionCtr );
                }
            }
        }
    }

    /**
     * Returns whether the given element is a select, an input or a button tag
     * or not.
     *
     * @param element  the element to query
     * @return         whether it's an input element or not
     */
    private boolean isInputElement( Element element )
    {
        if ( element.getNodeName().equalsIgnoreCase( "select" ) ||
                element.getNodeName().equalsIgnoreCase( "input" ) ||
                element.getNodeName().equalsIgnoreCase( "button" ) )
            return true;
        else
            return false;
    }

    /**
     * Searches the DOM tree downwards for a node named "table", starting with
     * the given node's first child node. If such a node is found, true is
     * returned, else false.
     *
     * @param node  the node to search
     * @return      whether the node contains a tag named "table"
     *      (case-insensitive search)
     */
    private boolean containsTableNode( Node node )
    {
        boolean bReturnValue = false;
        for ( Node child = node.getFirstChild(); child != null; child = child.getNextSibling() )
        {
            if ( child.getNodeName().equalsIgnoreCase( "table" ) )
                return true;

            if ( bReturnValue = containsTableNode( child ) )
                break;
        }
        return bReturnValue;
    }

    /**
     * Calculates the current index within the allRepeatedElements, i.e. which
     * of the repeated elements is in turn to be repeated.
     *
     * @param iRepetitionCtr       how many repetition cycles have been run on
     *      the current ID node yet?
     * @param allRepeatedElements  the list of RepeatedElement instances
     * @return                     the index within the allRepeatedElements
     *      (index starts with 0)
     */
    private int getIndex( int iRepetitionCtr, RepeatedElement[] allRepeatedElements )
    {
        return iRepetitionCtr % allRepeatedElements.length;
    }
}
