package org.alltimeflashdreamer.dimdom;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;
import org.xml.sax.*;

/**
 * Title: SiteTester<br>
 * Description: This class allows both a designer and a programmer to test
 * whether a html file describing a web site is acceptable for the DOM parser
 * and the needs of DimDom.<br>
 * It reads the command line parameters and parses the specified file
 * accordingly, printing out either the result output after a successful parsing
 * or the stack trace of the exception that was thrown during the processing.
 * <br>
 * Furthermore, SiteTester provides an installation-testing procedure, which may
 * be initiated using the word "test" as the first command line parameter. In
 * this case, SiteTester parses the hard-coded XHTML test string. If this
 * succeeds, the DimDom installation works correctly.
 *
 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

public class SiteTester
{
    private static boolean debug = false;

    /**
     * Tries to parse the file and throws the possibly upcoming exceptions to
     * the main method where they are printed to the screen.
     *
     * @param sFileName      the full path name of the file to be parsed or
     *      "test"
     * @exception Exception  any possibly upcoming exception
     */
    public SiteTester( String sFileName )
        throws Exception
    {
        //test mode?
        if ( sFileName.toLowerCase().equals( "test" ) )
        {
            runTestMode();
            return;
        }

        //normal file parsing mode
        TestDocument testDoc = new TestDocument( sFileName );
        if ( debug )
            System.out.println( testDoc.print() );

        //In case of success, write some output:
        System.out.println( "\nSuccessfully tested file " + sFileName + "!\n" );
        if ( debug )
        {
            System.out.println( "The file contains the following ID attribute values, placed within the " +
                    "specified HTML tag:" );
            Hashtable hIDNodeList = testDoc.hIDNodeList;
            Enumeration eKeys = hIDNodeList.keys();
            int iCtr = 0;
            while ( eKeys.hasMoreElements() )
            {
                String sKey = ( String )eKeys.nextElement();
                Element element = ( Element )hIDNodeList.get( sKey );
                System.out.println( "" + ( ++iCtr ) + ": " + sKey + " (" + element.getNodeName() + ")" );
            }
        }
    }

    /**
     * The main method parses the command line arguments and reacts accordingly.
     * <br>
     * The arguments are:
     * <ol>
     *   <li> the fully path and name of the file to be parsed for testing
     *   purposes (required, if a file shall be parsed)</li>
     *   <li> a boolean ("true" or "false") indicating whether debug information
     *   shall be printed or not (optional)</li>
     * </ol>
     * If the first argument is the word "test", the internal, very simple test
     * XHTML string is parsed, which is valid anyway. Should this throw an
     * error, something with your DimDom installation must be wrong.
     *
     * @param args  the command line arguments
     */
    public static void main( String[] args )
    {
        if ( isHelpRequired( args, System.out ) )
            return;

        String sFileName = args[0];

        //tests passed
        try
        {
            SiteTester siteTester = new SiteTester( sFileName );
        } catch ( Exception ex )
        {
            ex.printStackTrace();
        }
    }

    /**
     * Checks whether the user called this class with "-?", "/?", "-help" or
     * "/help" as the first command line parameter asking for usage help.<br>
     * If so, the help text is printed to the out and true is returned, else
     * nothing is done and false is returned.
     *
     * @param args  the command line arguments
     * @param out   where to print the output to
     * @return      whether the command line arguments are correct or not, i.e.
     *      help is required (leads to the printing of the usage text)
     */
    private static boolean isHelpRequired( String[] args, PrintStream out )
    {
        String sFirstArg = null;
        boolean bReturnValue = false;

        if ( ( args != null ) && ( args.length > 0 ) )
            sFirstArg = args[0].toLowerCase();

        if ( ( sFirstArg == null ) ||
                ( sFirstArg.equals( "-?" ) ) ||
                ( sFirstArg.equals( "/?" ) ) ||
                ( sFirstArg.equals( "-help" ) ) ||
                ( sFirstArg.equals( "/help" ) ) )
        {
            System.out.println( "Usage: java org.alltimeflashdreamer.dimdom.SiteTester filename [true | false]" );
            bReturnValue = true;
        }

        //check for debug setting
        if ( args.length > 1 )
        {
            try
            {
                debug = new Boolean( args[1].toLowerCase() ).booleanValue();
            } catch ( Exception ex )
            {
                bReturnValue = true;
            }
        }

        return bReturnValue;
    }

    /**
     * Runs the installation-testing mode.
     *
     * @exception Exception  any possibly upcoming exception
     */
    private void runTestMode()
        throws Exception
    {
        try
        {
            Document testDoc = new XHTMLParser().parse( DimDom.TESTDOCUMENT );
        } catch ( Exception ex )
        {
            System.out.println( "The parsing of the test document threw the following exception:" );
            ex.printStackTrace();
            System.out.println( "Please verify your DimDom installation." );
        }

        //In case of success, write some output:
        System.out.println( "Successfully run the test mode.\nYour DimDom installation is up and running." );
    }

//--------------- inner class TestDocument -----------------//
    /**
     * This class is a "specific implementation" for each XHTML file to be
     * tested.
     *
     * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
     *      Florian Steinsiepe</a>
     * @version   1.0
     */
    private class TestDocument extends SimpleDocument
    {
        /**
         * Constructor calls superclass' parseFile method using sFileName.
         *
         * @param sFileName            the full path name of the file to be
         *      parsed
         * @exception SAXException     thrown by the parser
         * @exception IOException      thrown by the parser
         * @exception DimDomException  thrown by the parser
         */
        private TestDocument( String sFileName )
            throws SAXException, IOException, DimDomException
        {
            this.parseFile( sFileName );
        }
    }
}
