package org.alltimeflashdreamer.dimdom;

import java.io.*;
import org.enhydra.xml.io.*;
import org.w3c.dom.Document;
import org.xml.sax.*;

/**
 * Title: XHTMLParser<br>
 * Description: Parses a File or an InputStream source and returns the Document
 * instance. The parser is the one of Enhydra's XMLC engine. Note: This is a
 * <i>validating</i> parser, therefore each input source needs a valid XHTML
 * doctype tag and must be fully XHTML compliant, which means that each tag must have
 * an ending tag and each attribute needs a value that is written between "" or
 * ''.<br>
 * Examples:
 * <ul>
 *   <li> &lt;br <b>/</b> &gt; or &lt;br&gt;<b>&lt;/br&gt;</b> </li>
 *   <li> &lt;img src="x.gif" alt="image" <b>/</b> &gt;</li>
 *   <li> &lt;option value="value" <b>selected="on"</b> &gt;Option Text<b>
 *   &lt;/option&gt;</li>
 * </ul>
 *
 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

public class XHTMLParser
{

    /** Empty constructor. */
    public XHTMLParser() { }

    /**
     * Parses the file and returns a Document object containing the DOM of the
     * file. <br>
     * The parsing itself is done with the validating Xerces XHTML parser.
     *
     * @param file              the file to parse
     * @return                  an org.w3c.dom.Document implementing object
     *      instance
     * @exception IOException   thrown by the DOMParser's parse method
     * @exception SAXException  thrown by the DOMParser's parse method
     */
    public Document parse( File file )
        throws IOException, SAXException
    {
        return parse( new FileInputStream( file ) );
    }

    /**
     * Parses the input stream and returns a Document object containing the DOM
     * of the stream.<br>
     * The parsing itself is done with the validating Xerces XHTML parser.
     *
     * @param stream            the InputStream to parse from
     * @return                  an org.w3c.dom.Document implementing object
     *      instance
     * @exception IOException   thrown by the DOMParser's parse method
     * @exception SAXException  thrown by the DOMParser's parse method
     */
    public Document parse( InputStream stream )
        throws IOException, SAXException
    {
        return parse( new InputSource( stream ) );
    }

    /**
     * Parses the input string and returns a Document object containing the DOM
     * of the content.<br>
     * The parsing itself is done with the validating Xerces XHTML parser.
     *
     * @param sXHTML            the validatable XHTML String to parse
     * @return                  an org.w3c.dom.Document implementing object
     *      instance
     * @exception IOException   thrown by the DOMParser's parse method
     * @exception SAXException  thrown by the DOMParser's parse method
     */
    public Document parse( String sXHTML )
        throws IOException, SAXException
    {
        return parse( new InputSource( new StringReader( sXHTML ) ) );
    }

    /**
     * Parses from a InputSource. This method effectively uses the XMLC
     * DOMParser, the others only transform their input source.
     *
     * @param source            the XMLC collection of input sources
     * @return                  the parsed Document instance
     * @exception IOException   thrown by the DOMParser's parse method
     * @exception SAXException  thrown by the DOMParser's parse method
     */
    private Document parse( InputSource source )
        throws IOException, SAXException
    {
        DOMParser parser = new DOMParser();
        parser.setValidation( true );
        return parser.parse( source );
    }
}
