package org.alltimeflashdreamer.dimdom.websitetester;

import org.alltimeflashdreamer.dimdom.*;

/**
 * Title: ComplexSiteProcessor<br>
 * Description: This is one of the reference implementations of the DimDom Website
 * Tester application. It is designed to fill the id nodes defined in the <i> default
 * </i> XHTML content of the text area on the first site. Therefore, if this XHTML
 * input changes, the behaviour of the engine may become inpredictable.<br>
 * Everything is done within the constructor <i>without arguments</i> , to allow
 * the XHTMLProcessorServlet to get this class via the newInstance method.<br>
 * The values to set are hard-coded.
 *
 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch"> Florian
 *      Steinsiepe</a>
 * @version   0.8.2
 */

public class ComplexSiteProcessor extends SimpleSiteData
{
    //some entries for the "places" select list
    private final static String[] PLACES = {"New York", "Berlin", "Venezia"};
    //some entries for the "times" select list
    private final static String[] TIMES = {"08:00", "12:13", "18.45", "23.59"};
    //some entries for the "events" select list (two-dimensional, as used if the URL parameter
    //value is not the same as the text shown within the list)
    private final static String[][] EVENTS = {{"1", "City Marathon"}, {"2", "Love Parade"}, {"3", "Biennale"}};

    //three names for the three repeated table rows
    private final static String[] NAMES = {"Jupiter", "Io", "Saturn"};
    //some image URL's for the images within a repeated table row
    private final static String[] IMAGES = {"http://www.jpl.nasa.gov/galileo/graphics/jupiterbox.jpg", "http://www.jpl.nasa.gov/galileo/graphics/io_pizza.jpg", "http://www.jpl.nasa.gov/galileo/graphics/saturn_vger2.gif"};
    //the name of the title of the attributes list
    private final static String ATTRIBTITLE = "Attributes";
    //three attribute lists for the three repeated table rows
    private final static String[][] ATTRIBS = {{"5th planet", "largest", "gas giant"}, {"moon of Jupiter", "life expected"}, {"6th planet", "many rings", "gas giant"}};

    /**
     * The constructor defines the commands for the id nodes defined in the template.
     * <br>
     * The template consists of two main parts: A form with three selection lists
     * and a table with a template table row that shall be filled and repeated several
     * times.<br>
     * The form's lists are filled dynamically, each with its specific selected entry,
     * and the three table row's cells are dynamized in the following order: name,
     * list of images, attribute title and attribute list.
     *
     * @exception DimDomException  the exception that tells the caller that the setting
     *      of some attribute values was illegal, i.e. the values haven't been XHTML
     *      compliant
     */
    public ComplexSiteProcessor()
        throws DimDomException
    {
        //call the empty constructor of the superclass
        super( "" );

        //add data to the "siteTitle" node
        createDimDomCommand( "siteTitle" ).addData( "ComplexSiteProcessor's dynamic output" );

		//set the common font attribute
		createDimDomCommand( "commonFont" ).setAttribute( "face", "Arial, Helvetica", true );

        //set the "selectForm"'s attributes (senseless action...)
        DimDomNode ddn = createDimDomCommand( "selectForm" );
        ddn.setAttribute( "action", "..." );
        ddn.setAttribute( "method", "post" );

        //add list of values to the "selectPlace" node and select the entry named "Berlin"
        createDimDomCommand( "selectPlace" ).addData( PLACES, "Berlin" );
        //add list of values to the "selectTime" node and select the third entry (id = 2)
        createDimDomCommand( "selectTime" ).addData( TIMES, 2 );
        //add list of values to the "selectEvent" node and select the second entry row (id = 1)
        createDimDomCommand( "selectEvent" ).addData( EVENTS, 1 );

        //loop three times over the repeated table row...
        for ( int i = 0; i < 3; i++ )
        {
            //add the name of the current table row
            createDimDomCommand( "repeatedTableRow" ).addData( NAMES[i] );

            //create a random number defining the amount of visible images (from 1 to 4)
            int iImgCount = ( int )( Math.random() * 3 ) + 1;

            //loop over the four image templates and either set their src attributes or
            //set them invisible, depending on the amount of visible images
            for ( int j = 0; j < 4; j++ )
            {
                if ( j < iImgCount )
                    createDimDomCommand( "repeatedTableRow" ).setAttribute( "src", IMAGES[i] );
                else
                    createDimDomCommand( "repeatedTableRow" ).setVisible( false );
            }

            //set the title of the attributes list
            createDimDomCommand( "repeatedTableRow" ).addData( ATTRIBTITLE );

            //create an empty command which repeats the "br" tag
            createDimDomCommand( "repeatedTableRow" );

            //set the list of attributes of the current table row
            createDimDomCommand( "repeatedTableRow" ).addData( ATTRIBS[i] );
        }
    }
}
