package org.alltimeflashdreamer.dimdom.websitetester;

import java.io.*;
import java.util.Date;
import javax.servlet.ServletException;
import org.alltimeflashdreamer.dimdom.*;

/**
 * Title: ErrorSiteProcessor<br>
 * Description: This class generates the error output site, in case an error
 * came up somewhere and needs to be shown to the user. Therefore, it passes the
 * error site template (which is hard-coded string) and the appropriately filled
 * XHTMLSiteData instance to the DimDom engine which then produces the error
 * site output.
 *
 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

class ErrorSiteProcessor
{
    //error codes
    /** code describing the error during finding/reading of the property file */
    final static int PROPERTYFILEERROR = 0;
    /**
     * code describing that no XHTML input string was found/given through the
     * tester web site
     */
    final static int MISSINGINPUTERROR = 1;
    /** code describing the error during parsing of the XHTML source/input */
    final static int INPUTPARSERERROR = 2;
    /**
     * code describing that an error occurred during fill-in of the dynamic data
     * into the DimDom command objects
     */
    final static int SITEDATAERROR = 3;
    /**
     * code describing that an error occurred during printing out of the
     * dynamically updated Document object
     */
    final static int SITEPRINTERROR = 4;
    /** code describing that the given processor class wasn't found */
    final static int MISSINGPROCESSORCLASSERROR = 5;

    //error messages, with above defined indexes
    private final static String[]
            ERRORMESSAGES = {
            "The property file couldn't be found, neither as a servlet parameter nor in the servlet's " +
            " web directory!\nPlease verify your installation.",
            "The URL input parameter for the XHTML string to parse is missing or empty. Either the source " +
            "site has been changed or the text input was empty.\nPlease verify your installation or enter a XHTML string.",
            "The XHTML parser threw an exception! This means that the given input string is not a valid " +
            "XHTML string as described on the site!\nPlease go back and verify your input.",
            "An exception occurred during fill-in of the dynamic values into the DimDom command objects.\n" +
            "Please make sure that you didn't modify the code (if you're using the SampleSiteProcessor class) " +
            "or that your own processor class fits the needs of the given input. If it's none of the mentioned cases, " +
            "the attribute values that shall be added to the DOM tree aren't XHTML compliant.",
            "The DimDom class threw an exception while printing out the dynamically filled DOM tree!\n" +
            "This means that either your installation is wrong or the SampleSiteProcessor code has been " +
            "modified illegally or your own processor class isn't valid.",
            "The given processor class was not found!\nPlease make sure that this class is available for " +
            "your servlet runner."
            };

    //DimDom node id names
    private final static String HELPTEXT = "helpText";
    private final static String ERRORTEXT = "errorText";
    private final static String FOOTERTEXT = "footerText";

    //XHTML template (string)
    private final static String
            XHTML = "<html>\n" +
            "\t<head>\n" +
            "\t\t<meta http-equiv=\"content-type\" content=\"text/html;charset=iso-8859-1\" />\n" +
            "\t\t<title>DimDom Website Tester: Error site</title>\n" +
            "\t</head>\n\n" +
            "\t<body bgcolor=\"#ffffff\" text=\"#000033\" link=\"#000033\" alink=\"#000033\" vlink=\"#000033\" leftmargin=\"10\" marginwidth=\"10\" topmargin=\"10\" marginheight=\"10\">\n" +
            "\t\t<div align=\"center\">\n" +
            "\t\t\t<h2><font face=\"Arial,Helvetica,Geneva,Swiss,SunSans-Regular\">DimDom Website Tester: Error site</font></h2>\n" +
            "\t\t\t<hr />\n" +
            "\t\t</div>\n" +
            "\t\t<p>The following error occurred while parsing the given input string: </p>\n" +
            "\t\t<p id=\"" + HELPTEXT + "\">helptexthelptexthelptext</p>\n" +
            "\t\t<p id=\"" + ERRORTEXT + "\">errortexterrortexterrortext</p>\n" +
            "\t\t<hr />\n" +
            "\t\t<p id=\"" + FOOTERTEXT + "\"><font size=\"-2\">footertextfootertext</font></p>\n" +
            "\t</body>\n" +
            "\t</html>";

    //the currently required error code to handle
    /** Description of the Field */
    int iErrorCode;
    //the current exception to be commented (may be null)
    /** Description of the Field */
    Exception ex;


    /**
     * Simply stores the given parameters, for later processing.
     *
     * @param iErrorCode  which error shall be shown?
     * @param ex          the exception that occurred (may be null)
     */
    ErrorSiteProcessor( int iErrorCode, Exception ex )
    {
        this.iErrorCode = iErrorCode;
        this.ex = ex;
    }

    /**
     * Builds the error site output and calls the print method of the DimDom
     * class to get the fully formatted, dynamically filled XHTML string.<br>
     * If somehow this class itself runs into troubles, it wraps them into an
     * IOException.
     *
     * @return                   the fully formatted XHTML string
     * @throws ServletException  if the processor itself suffers from some
     *      exception...
     */
    String print()
        throws ServletException
    {
        SimpleSiteData siteData = new SimpleSiteData( XHTML );
        String sErrorText = "";
        if ( ex == null )
            sErrorText = "(no related exception)";
        else
        {
            StringWriter sw = new StringWriter();
            ex.printStackTrace( new PrintWriter( sw ) );
            sErrorText = sw.toString();
        }
        try
        {
            siteData.createDimDomCommand( HELPTEXT ).addData( ERRORMESSAGES[iErrorCode], true );
            siteData.createDimDomCommand( ERRORTEXT ).addData( sErrorText );
            String sFooterText = "DimDom Error Site Processing: " +
                    new Date().toString();
            siteData.createDimDomCommand( FOOTERTEXT ).addData( sFooterText );
            return DimDom.print( siteData );
        } catch ( Exception ex )
        {
            //wrap each upcoming exception in a ServletException
            throw new ServletException( "Error output site couldn't be created!", ex );
        }
    }
}
