package org.alltimeflashdreamer.dimdom.websitetester;

import java.io.*;
import java.net.*;
import java.util.Properties;
import javax.servlet.*;
import javax.servlet.http.*;

/**
 * Title: PropertyHandler<br>
 * Description: Simple value handler class for the various property file parameters.
 *
 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */
class PropertyHandler
{
    /**
     * the name of the servlet parameter describing the property file, if
     * defined; this is useful if one would like to hide this file and updates
     * the web.xml file to allow the servlet to find the property elsewhere
     */
    private final static String PROPERTYFILE_PARAMETER = "DimDom_WebsiteTester";
    /** the name of the file to read if the servlet parameter isn't found */
    private final static String PROPERTYFILE = PROPERTYFILE_PARAMETER + ".properties";

    //property file key constants
    private final static String[]
            PROPERTIES = {"DEFAULT_PROCESSOR_CLASS", "XML_TAG", "DOCTYPE_TAG",
            "XHTMLPARSERSERVLETSITE"};
    private final static int DEFAULT_PROCESSOR_CLASS = 0;
    private final static int XML_TAG = 1;
    private final static int DOCTYPE_TAG = 2;
    private final static int XHTMLPARSERSERVLETSITE = 3;

    //other constants
    private final static String NEWLINE = System.getProperty( "line.separator" );

    //class fields
    private Properties props = new Properties();
    private boolean bDidReadProperties = false;
    private Exception exception;


    /**
     * Reads the property file from the various input source and stores the
     * state, i.e. whether this succeeded or not.
     *
     * @param req      the request that lead to the initiation of this class
     * @param context  the servlet's context, to read a servlet parameter from
     */
    PropertyHandler( HttpServletRequest req, ServletContext context )
    {
        readPropertyFile( req, context );
    }

    /**
     * Tries to read the property file from either the servlet config or from
     * the servlet's web directory. The method registeres whether this succeeded
     * or not, for later querying.
     *
     * @param req      the request that lead to the initiation of this class
     * @param context  the servlet's context, to read a servlet parameter from
     */
    private void readPropertyFile( HttpServletRequest req, ServletContext context )
    {
        //try to get the servlet parameter for the property file
        String sPropertyFile = ( String )context.getAttribute( PROPERTYFILE_PARAMETER );
        if ( sPropertyFile != null )
        {
            try
            {
                props.load( new FileInputStream( sPropertyFile ) );
                bDidReadProperties = true;
            } catch ( Exception ex )
            {}
        }

        //if not found, try to load the property file as an URL input stream
        if ( !hasReadingBeenSuccessful() )
        {
            sPropertyFile = HttpUtils.getRequestURL( req ).toString();

            //take everything until the last "/", cut "/servlet", if the
            //servlet runs on Tomcat, and try to build the URL of the property file
            sPropertyFile = sPropertyFile.substring( 0, sPropertyFile.lastIndexOf( '/' ) );
            if ( sPropertyFile.endsWith( "/servlet" ) )
                sPropertyFile = sPropertyFile.substring( 0, sPropertyFile.length() - "/servlet".length() );
            String sPort = "" + req.getServerPort();
            if ( sPropertyFile.indexOf( sPort ) >= 0 )
                sPropertyFile = sPropertyFile.substring( sPropertyFile.indexOf( sPort ) + sPort.length() );
            if ( !sPropertyFile.endsWith( "/" ) )
                sPropertyFile += "/";
            try
            {
                props.load( new URL( "http", req.getServerName(),
                        req.getServerPort(), sPropertyFile + PROPERTYFILE ).openStream() );
                bDidReadProperties = true;
            } catch ( Exception ex )
            {
                exception = ex;
            }
        }

        //verify existence of parameters, if they were read
        if ( hasReadingBeenSuccessful() )
        {
            for ( int i = 0; i < PROPERTIES.length; i++ )
            {
                String sProperty = getProperty( i );
                if ( sProperty == null )
                {
                    bDidReadProperties = false;
                    exception = new NoSuchFieldException( "Parameter \"" + PROPERTIES[i] +
                            "\" is missing in the property file!" );
                    break;
                }

                //special checks
                switch ( i )
                {
                    case XHTMLPARSERSERVLETSITE:
                        try
                        {
                            URL dummy = new URL( sProperty );
                        } catch ( MalformedURLException mue )
                        {
                            bDidReadProperties = false;
                            exception = mue;
                        }
                        break;
                    default:
                        break;
                }
            }
        }
    }

    /**
     * Returns whether the property file was found and loaded or not. If this
     * method returns false, the getException method returns the exception that
     * occurred.
     *
     * @return   everything fine?
     */
    boolean hasReadingBeenSuccessful()
    {
        return bDidReadProperties;
    }

    /**
     * Returns the exception that came up or null.
     *
     * @return   not null only if the hasReadingBeenSuccessful method returns
     *      true
     */
    Exception getException()
    {
        return exception;
    }

    /**
     * Returns two full line, one for the xml and one for the doctype tags, both
     * ending with a line separator.
     *
     * @return   xml-tag + newline + doctype-tag + newline
     */
    String getXMLTags()
    {
        return getProperty( XML_TAG ) + NEWLINE +
                getProperty( DOCTYPE_TAG ) + NEWLINE;
    }

    /**
     * Returns the URL of the template file of the parser output site. This URL
     * has been validated during reading of the property file.
     *
     * @return   the URL, ready to open as an input stream
     */
    URL getParserServletSiteURL()
    {
        try
        {
            return new URL( getProperty( XHTMLPARSERSERVLETSITE ) );
        } catch ( MalformedURLException mue )
        {
            //is checked, must never happen*
            return null;
        }
    }

    /**
     * Returns the name of the default processor class, as found in the property
     * file.
     *
     * @return   the full class name of the input processor class
     */
    String getProcessorClassName()
    {
        return getProperty( DEFAULT_PROCESSOR_CLASS );
    }

    /**
     * Simple accessor for a property specified by its id.
     *
     * @param iWhichOne  the id of the property
     * @return           the value of the given property id
     */
    private String getProperty( int iWhichOne )
    {
        return props.getProperty( PROPERTIES[iWhichOne] );
    }
}
