package org.alltimeflashdreamer.dimdom.websitetester;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.w3c.dom.*;
import org.alltimeflashdreamer.dimdom.*;

/**
 * Title: XHTMLParserServlet<br>
 * Description: This servlet parses the XHTML string from URL input and tries to
 * build a Document instance out of it. If this succeeds, the list of recognized
 * DimDomNodes is shown (of course using a DimDom-processed template), else the
 * catched exception's stack trace is shown.
 *
 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
 *      Florian Steinsiepe</a>
 * @version   0.8.2
 */

public class XHTMLParserServlet extends AbstractXHTMLServlet
{
    /**
     * Reads the XHTML string from the URL and passes it to the XHTMLParser
     * class of DimDom. If the parser fails to build a Document out of this
     * string, it shows the processed error site. If it succeeds, it fills the
     * success site with the dynamically recognized DimDom-relevant data from
     * the template and allows the user to view the reflection output.<br>
     * The various input parameters are read from the
     * DimDom_WebsiteTester.properties file. The method first asks for a servlet
     * parameter named "DimDom_WebsiteTester" (case-sensitive). If this
     * parameter is found, it reads the file mentioned in this parameter, else
     * it tries to read it from the servlet's application directory. If this
     * fails too, an exception is thrown.
     *
     * @param req                the request object to read URL stuff from
     * @param res                the response object to send stuff to
     * @throws ServletException  as defined in the superclass
     * @throws IOException       as defined in the superclass
     */
    public void doPost( HttpServletRequest req, HttpServletResponse res )
        throws ServletException, IOException
    {
        //read the property file
        PropertyHandler propsHandler = new PropertyHandler( req, this.getServletContext() );
        if ( !propsHandler.hasReadingBeenSuccessful() )
        {
            showErrorSite( ErrorSiteProcessor.PROPERTYFILEERROR, res, propsHandler.getException() );
            return;
        }

        //get the URL input parameter value and add the xml and the doctype tags to it.
        String sXHTMLInput = req.getParameter( "dimdomXHTMLInput" );
        if ( ( sXHTMLInput == null ) || ( sXHTMLInput.equals( "" ) ) )
        {
            showErrorSite( ErrorSiteProcessor.MISSINGINPUTERROR, res, null );
            return;
        }
        sXHTMLInput = propsHandler.getXMLTags() + sXHTMLInput;

        //parse the input, if DimDom doesn't know the identifier yet
        XHTMLDocumentImpl docImpl = null;
        try
        {
            if ( ( docImpl = ( XHTMLDocumentImpl )DimDom.getDimDomDocument( sXHTMLInput ) ) == null )
            {
                docImpl = new XHTMLDocumentImpl( new XHTMLParser().parse( sXHTMLInput ) );
                DimDom.addDimDomDocument( sXHTMLInput, docImpl );
            }
        } catch ( ClassCastException cce )
        {
            try
            {
                docImpl = new XHTMLDocumentImpl( new XHTMLParser().parse( sXHTMLInput ) );
                DimDom.addDimDomDocument( sXHTMLInput, docImpl );
            } catch ( Exception ex )
            {
                showErrorSite( ErrorSiteProcessor.INPUTPARSERERROR, res, ex );
                return;
            }
        } catch ( Exception ex )
        {
            showErrorSite( ErrorSiteProcessor.INPUTPARSERERROR, res, ex );
            return;
        }

        //prepare the input stream so that its a string and it starts with
        //the specified XML tags (from the property file)
        URL templateURL = propsHandler.getParserServletSiteURL();
        BufferedReader br = new BufferedReader( new InputStreamReader( templateURL.openStream() ) );
        StringBuffer sbTemplateInput = new StringBuffer( propsHandler.getXMLTags() );
        String sNextLine = "";
        while ( ( sNextLine = br.readLine() ) != null )
            sbTemplateInput.append( sNextLine );

        //fill the DimDom command objects and set the XHTMLSiteData to docImpl
        SimpleSiteData siteData = new SimpleSiteData( sbTemplateInput.toString() );
        try
        {
            //normal DimDomNodes
            DimDomNodeImpl ddni = siteData.createDimDomCommand( "normalDimDomNodes" );
            Hashtable hIDNodes = docImpl.getIDNodeList();
            Enumeration eIDs = hIDNodes.keys();
            while ( eIDs.hasMoreElements() )
            {
                String sNextID = ( String )eIDs.nextElement();
                ddni.addData( sNextID + ": " + ( ( Node )hIDNodes.get( sNextID ) ).getNodeName() );
            }

            //repeated DimDomNodes
            ddni = siteData.createDimDomCommand( "repeatedDimDomNodes" );
            Hashtable hRepElems = docImpl.getRepeatedElements();
            eIDs = hRepElems.keys();
            while ( eIDs.hasMoreElements() )
            {
                String sNextID = ( String )eIDs.nextElement();
                ddni.addData( sNextID + ": " + ( String )hRepElems.get( sNextID ) );
            }

            //processor class
            ddni = siteData.createDimDomCommand( "processorClass" );
            ddni.setAttribute( "name", "processorClass" );
            ddni.setAttribute( "value", propsHandler.getProcessorClassName() );

            //footer text
            siteData.createDimDomCommand( "footerText" ).addData(
                    "DimDom Parser Site Processing: " + new Date().toString() );

            //pass the commands to the XHTMLSiteData
            docImpl.setSiteData( siteData );
        } catch ( Exception ex )
        {
            showErrorSite( ErrorSiteProcessor.SITEDATAERROR, res, ex );
            return;
        }

        //pass the XHTMLSiteData to DimDom and print it out
        printSite( res, siteData );
    }

//--------------------------- inner class XHTMLDocumentImpl ---------------------------------//
    /**
     * Title: XHTMLDocumentImpl<br>
     * Description: This very simple extension of AbstractXHTMLDocument allows
     * the use of the methods of this abstract class.
	 *
	 * @author    Copyright (c) 2002 by <a href="mailto:florian@steinsiepe.ch">
	 *      Florian Steinsiepe</a>
	 * @version   1.0
     */
    private class XHTMLDocumentImpl extends AbstractXHTMLDocument
    {
        private Document doc;
        private Hashtable hIDNodes;

        /**
         * The constructor tries to build the reflection information. If this
         * fails, the upcoming exception is thrown.<br>
         * If the building was successful, the list of id nodes and the list of
         * repeated elements is built, to allow to show these lists to the user.
         *
         * @param doc            the Document representing the template
         * @throws Exception     if the parsing or the building of the lists
         *      failed somehow
         */
        private XHTMLDocumentImpl( Document doc )
            throws Exception
        {
            this.doc = doc;
            hIDNodes = this.buildListOfIDNodes( doc );
            this.buildMicroTemplateNodeList( hIDNodes );
        }

        /**
         * Returns the list of id nodes found in the input. This list consists
         * of key/value pairs (key: id value, value: underlying Element object).
         *
         * @return   the list of ID nodes (a clone, for thread-safety reasons)
         */
        private Hashtable getIDNodeList()
        {
            return ( Hashtable )this.hIDNodes.clone();
        }

        /**
         * Returns the list of repeated nodes found in the input. This list
         * consists of key/value pairs (key: id value, value: maximum amount of
         * repetitions, as a descriptive string). The id is listed as many times
         * as there are such nodes.
         *
         * @return   the list of repeated nodes (built from a clone, for
         *      thread-safety reasons)
         */
        private Hashtable getRepeatedElements()
        {
            Hashtable hReturnValue = new Hashtable();
            Enumeration eIDs = this.hMicroTemplateNodes.keys();
            while ( eIDs.hasMoreElements() )
            {
                String sNextID = ( String )eIDs.nextElement();
                RepeatedElement[] repElems = ( RepeatedElement[] )this.hMicroTemplateNodes.get( sNextID );
                for ( int i = 0; ( repElems != null ) && ( i < repElems.length ); i++ )
                {
                    String sMaxRepetitions = repElems[i].getRepeatedNode().getNodeName() +
                            ", " + ( ( repElems[i].getMaxRepetitionCount() == -1 )
                             ? "unlimited" : "" + repElems[i].getMaxRepetitionCount() ) + " repetition(s)";
                    hReturnValue.put( sNextID + " (" + ( i + 1 ) + ")", sMaxRepetitions );
                }
            }

            return hReturnValue;
        }

        /**
         * (default implementation of abstract method, throws an
         * UnsupportedOperationException)
         *
         * @param sNodeIdentifier      the id attribute value of the node to be
         *      processed
         * @param element              the current node to be processed, for
         *      deciding how to behave
         * @return                     the next repeated element, i.e. a clone
         *      that is already appended/inserted correctly
         * @exception DimDomException  wrapper for unexpected errors
         */
        protected Element getNextRepeatedElement( String sNodeIdentifier, Element element )
            throws DimDomException
        {
            throw new UnsupportedOperationException();
        }
    }
}
